"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from dataclasses import dataclass

from cyclonedds.idl import IdlStruct
from cyclonedds.idl.types import array, sequence
from cyclonedds.idl.annotations import key


@dataclass
class Integer(IdlStruct):
    seq: int
    keyval: int
    key("seq")
    key("keyval")

    @classmethod
    def postfix(cls):
        return "integer"


@dataclass
class String(IdlStruct):
    seq: int
    keyval: str
    key("seq")
    key("keyval")

    @classmethod
    def postfix(cls):
        return "string"


@dataclass
class IntArray(IdlStruct):
    seq: int
    keyval: array[int, 3]
    key("seq")
    key("keyval")

    @staticmethod
    def size():
        return 3

    @classmethod
    def postfix(cls):
        return "int_array"


@dataclass
class StrArray(IdlStruct):
    seq: int
    keyval: array[str, 5]
    key("seq")

    @staticmethod
    def size():
        return 5

    @classmethod
    def postfix(cls):
        return "str_array"


@dataclass
class IntSequence(IdlStruct):
    seq: int
    keyval: sequence[int]
    key("seq")

    @classmethod
    def postfix(cls):
        return "int_sequence"


@dataclass
class StrSequence(IdlStruct):
    seq: int
    keyval: sequence[str, 100]  # max 100 string elements
    key("seq")

    @classmethod
    def postfix(cls):
        return "str_sequence"


datatypes = [Integer, String, IntArray, StrArray, IntSequence, StrSequence]



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from typing import Optional, Union, Generic, TypeVar, List, TYPE_CHECKING
import ctypes as ct
import uuid

from .internal import c_call, dds_c_t
from .core import Entity, DDSException, Listener
from .domain import DomainParticipant
from .topic import Topic
from .qos import _CQos, Qos, LimitedScopeQos, PublisherQos, DataWriterQos
from .builtin_types import DcpsEndpoint, endpoint_constructor, cqos_to_qos

from cyclonedds._clayer import ddspy_write, ddspy_write_ts, ddspy_dispose, ddspy_writedispose, ddspy_writedispose_ts, \
    ddspy_dispose_handle, ddspy_dispose_handle_ts, ddspy_register_instance, ddspy_unregister_instance,   \
    ddspy_unregister_instance_handle, ddspy_unregister_instance_ts, ddspy_unregister_instance_handle_ts, \
    ddspy_lookup_instance, ddspy_dispose_ts, ddspy_get_matched_subscription_data


if TYPE_CHECKING:
    import cyclonedds


class Publisher(Entity):
    def __init__(
            self,
            domain_participant: DomainParticipant,
            qos: Optional[Qos] = None,
            listener: Optional[Listener] = None):
        if not isinstance(domain_participant, DomainParticipant):
            raise TypeError(f"{domain_participant} is not a cyclonedds.domain.DomainParticipant.")

        if qos is not None:
            if isinstance(qos, LimitedScopeQos) and not isinstance(qos, PublisherQos):
                raise TypeError(f"{qos} is not appropriate for a Publisher")
            elif not isinstance(qos, Qos):
                raise TypeError(f"{qos} is not a valid qos object")

        if listener is not None:
            if not isinstance(listener, Listener):
                raise TypeError(f"{listener} is not a valid listener object.")

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        try:
            super().__init__(
                self._create_publisher(
                    domain_participant._ref,
                    cqos,
                    listener._ref if listener else None
                ),
                listener=listener
            )
        finally:
            if cqos:
                _CQos.cqos_destroy(cqos)

        self._keepalive_entities = [self.participant]

    def suspend(self):
        ret = self._suspend(self._ref)
        if ret == 0:
            return
        raise DDSException(ret, f"Occurred while suspending {repr(self)}")

    def resume(self):
        ret = self._resume(self._ref)
        if ret == 0:
            return
        raise DDSException(ret, f"Occurred while resuming {repr(self)}")

    def wait_for_acks(self, timeout: int):
        """
        This operation blocks the calling thread until either all data written by the publisher
        or writer is acknowledged by all matched reliable reader entities, or else the duration
        specified by the timeout parameter elapses, whichever happens first.

        Parameters
        ----------
        timeout
            The maximum number of nanoseconds to wait. Use the function :func:`duration<cyclonedds.util.duration>`
            to write that in a human readable format.
        """
        ret = self._wait_for_acks(self._ref, timeout)
        if ret == 0:
            return True
        elif ret == DDSException.DDS_RETCODE_TIMEOUT:
            return False
        raise DDSException(ret, f"Occurred while waiting for acks from {repr(self)}")

    @c_call("dds_create_publisher")
    def _create_publisher(self, domain_participant: dds_c_t.entity, qos: dds_c_t.qos_p,
                          listener: dds_c_t.listener_p) -> dds_c_t.entity:
        pass

    @c_call("dds_suspend")
    def _suspend(self, publisher: dds_c_t.entity) -> dds_c_t.returnv:
        pass

    @c_call("dds_resume")
    def _resume(self, publisher: dds_c_t.entity) -> dds_c_t.returnv:
        pass

    @c_call("dds_wait_for_acks")
    def _wait_for_acks(self, publisher: dds_c_t.entity, timeout: dds_c_t.duration) -> dds_c_t.returnv:
        pass


_T = TypeVar('_T')

class DataWriter(Entity, Generic[_T]):
    def __init__(self,
                 publisher_or_participant: Union[DomainParticipant, Publisher],
                 topic: Topic[_T],
                 qos: Optional[Qos] = None,
                 listener: Optional[Listener] = None):
        if not isinstance(publisher_or_participant, (DomainParticipant, Publisher)):
            raise TypeError(f"{publisher_or_participant} is not a cyclonedds.domain.DomainParticipant"
                            " or cyclonedds.pub.Publisher.")

        if not isinstance(topic, Topic):
            raise TypeError(f"{topic} is not a cyclonedds.topic.Topic.")

        if qos is not None:
            if isinstance(qos, LimitedScopeQos) and not isinstance(qos, DataWriterQos):
                raise TypeError(f"{qos} is not appropriate for a DataWriter")
            elif not isinstance(qos, Qos):
                raise TypeError(f"{qos} is not a valid qos object")

        cqos = _CQos.qos_to_cqos(qos) if qos else None
        try:
            super().__init__(
                self._create_writer(
                    publisher_or_participant._ref,
                    topic._ref,
                    cqos,
                    listener._ref if listener else None
                ),
                listener=listener
            )
        finally:
            if cqos:
                _CQos.cqos_destroy(cqos)

        self._topic = topic
        self.data_type = topic.data_type
        self._keepalive_entities = [self.publisher, self.topic]
        self._constructor = None

        cqos = _CQos.cqos_create()
        ret = self._get_qos(self._ref, cqos)
        if ret == 0:
            data_repr_policy = _CQos._get_p_datarepresentation(cqos)
            if data_repr_policy is None or (data_repr_policy.use_cdrv0_representation and data_repr_policy.use_xcdrv2_representation):
                self._use_version_2 = None  # Use whatever is native to the datatype
            elif data_repr_policy.use_xcdrv2_representation:
                self._use_version_2 = True
            else:
                self._use_version_2 = False
        _CQos.cqos_destroy(cqos)

    @property
    def topic(self) -> Topic[_T]:
        return self._topic

    def write(self, sample: _T, timestamp: Optional[int] = None):
        """
        Parameters
        ----------
        sample
            The sample to write
        timestamp
            The sample's source_timestamp (in nanoseconds since the UNIX Epoch)
        """
        if not isinstance(sample, self.data_type):
            raise TypeError(f"{sample} is not of type {self.data_type}")

        ser = sample.serialize(use_version_2=self._use_version_2)
        ser = ser.ljust((len(ser) + 4 - 1) & ~(4 - 1), b'\0')

        if timestamp is not None:
            ret = ddspy_write_ts(self._ref, ser, timestamp)
        else:
            ret = ddspy_write(self._ref, ser)

        if ret < 0:
            raise DDSException(ret, f"Occurred while writing sample in {repr(self)}")

    def write_dispose(self, sample: _T, timestamp: Optional[int] = None):
        """
        Similar to :func:`write` but also marks the sample for disposal by setting its
        :class:`InstanceState<cyclonedds.core.InstanceState>` to `NotAliveDisposed`.

        Parameters
        ----------
        sample
            The sample to dispose
        timestamp
            The sample's source_timestamp (in nanoseconds since the UNIX Epoch)
        """
        ser = sample.serialize(use_version_2=self._use_version_2)
        ser = ser.ljust((len(ser) + 4 - 1) & ~(4 - 1), b'\0')

        if timestamp is not None:
            ret = ddspy_writedispose_ts(self._ref, ser, timestamp)
        else:
            ret = ddspy_writedispose(self._ref, ser)

        if ret < 0:
            raise DDSException(ret, f"Occurred while writedisposing sample in {repr(self)}")

    def dispose(self, sample: _T, timestamp: Optional[int] = None):
        """
        Marks the sample for disposal by setting its :class:`InstanceState<cyclonedds.core.InstanceState>` to
        `NotAliveDisposed`.

        Parameters
        ----------
        sample
            The sample to dispose
        timestamp
            The sample's source_timestamp (in nanoseconds since the UNIX Epoch)
        """
        ser = sample.serialize_key(use_version_2=self._use_version_2)
        ser = ser.ljust((len(ser) + 4 - 1) & ~(4 - 1), b'\0')

        if timestamp is not None:
            ret = ddspy_dispose_ts(self._ref, ser, timestamp)
        else:
            ret = ddspy_dispose(self._ref, ser)

        if ret < 0:
            raise DDSException(ret, f"Occurred while disposing in {repr(self)}")

    def dispose_instance_handle(self, handle: int, timestamp: Optional[int] = None):
        """
        Marks the instance and all samples associated wiht the given handle for disposal by setting their
        :class:`InstanceState<cyclonedds.core.InstanceState>` to `NotAliveDisposed`.

        Parameters
        ----------
        handle
            An instance handle received from :func:`register_instance` or :func:`lookup_instance`.
        timestamp
            The instance's source_timestamp (in nanoseconds since the UNIX Epoch)
        """
        if timestamp is not None:
            ret = ddspy_dispose_handle_ts(self._ref, handle, timestamp)
        else:
            ret = ddspy_dispose_handle(self._ref, handle)

        if ret < 0:
            raise DDSException(ret, f"Occurred while disposing in {repr(self)}")

    def register_instance(self, sample: _T) -> int:
        ser = sample.serialize_key(use_version_2=self._use_version_2)
        ser = ser.ljust((len(ser) + 4 - 1) & ~(4 - 1), b'\0')

        ret = ddspy_register_instance(self._ref, ser)
        if ret < 0:
            raise DDSException(ret, f"Occurred while registering instance in {repr(self)}")
        return ret

    def unregister_instance(self, sample: _T, timestamp: Optional[int] = None):
        """
        Parameters
        ----------
        sample
            The sample to unregister
        timestamp
            The timestamp used at registration (in nanoseconds since the UNIX Epoch)
        """
        ser = sample.serialize_key(use_version_2=self._use_version_2)
        ser = ser.ljust((len(ser) + 4 - 1) & ~(4 - 1), b'\0')

        if timestamp is not None:
            ret = ddspy_unregister_instance_ts(self._ref, ser, timestamp)
        else:
            ret = ddspy_unregister_instance(self._ref, ser)

        if ret < 0:
            raise DDSException(ret, f"Occurred while unregistering instance in {repr(self)}")

    def unregister_instance_handle(self, handle: int, timestamp: Optional[int] = None):
        """
        Parameters
        ----------
        handle
            An instance handle received from :func:`register_instance` or :func:`lookup_instance`.
        timestamp
            The timestamp used at registration (in nanoseconds since the UNIX Epoch)
        """
        if timestamp is not None:
            ret = ddspy_unregister_instance_handle_ts(self._ref, handle, timestamp)
        else:
            ret = ddspy_unregister_instance_handle(self._ref, handle)

        if ret < 0:
            raise DDSException(ret, f"Occurred while unregistering instance handle n {repr(self)}")

    def wait_for_acks(self, timeout: int) -> bool:
        """
        This operation blocks the calling thread until either all data written by the publisher
        or writer is acknowledged by all matched reliable reader entities, or else the duration
        specified by the timeout parameter elapses, whichever happens first.

        Parameters
        ----------
        timeout
            The maximum number of nanoseconds to wait. Use the function :func:`duration<cyclonedds.util.duration>`
            to write that in a human readable format.
        """
        ret = self._wait_for_acks(self._ref, timeout)
        if ret == 0:
            return True
        elif ret == Exception.DDS_RETCODE_TIMEOUT:
            return False
        raise DDSException(ret, f"Occurred while waiting for acks from {repr(self)}")

    def lookup_instance(self, sample: _T) -> Optional[int]:
        """
        This operation takes a sample and returns an instance handle to be used for subsequent operations.
        """
        ser = sample.serialize_key(use_version_2=self._use_version_2)
        ser = ser.ljust((len(ser) + 4 - 1) & ~(4 - 1), b'\0')

        ret = ddspy_lookup_instance(self._ref, ser)
        if ret < 0:
            raise DDSException(ret, f"Occurred while lookup up instance from {repr(self)}")
        if ret == 0:
            return None
        return ret

    def get_matched_subscriptions(self) -> List[int]:
        """Get instance handles of the data readers matching a writer.

        Raises
        ------
            DDSException: When the number of matching readers < 0.

        Returns
        -------
        List[int]:
            A list of instance handles of the matching data readers.
        """
        num_matched_sub = self._get_matched_subscriptions(self._ref, None, 0)
        if num_matched_sub < 0:
            raise DDSException(num_matched_sub, f"Occurred when getting the number of matched subscriptions of {repr(self)}")
        if num_matched_sub == 0:
            return []

        matched_sub_list = (dds_c_t.instance_handle * int(num_matched_sub))()
        matched_sub_list_pt = ct.cast(matched_sub_list, ct.POINTER(dds_c_t.instance_handle))

        ret = self._get_matched_subscriptions(self._ref, matched_sub_list_pt, num_matched_sub)
        if ret >= 0:
            return [matched_sub_list[i] for i in range(ret)]

        raise DDSException(ret, f"Occurred when getting the matched subscriptions of {repr(self)}")

    matched_sub = property(get_matched_subscriptions)

    def get_matched_subscription_data(self, handle) -> Optional['cyclonedds.builtin.DcpsEndpoint']:
        """Get a description of a reader matched with the provided writer

        Parameters
        ----------
        handle: Int
            The instance handle of a reader.

        Returns
        -------
        DcpsEndpoint:
            The sample of the DcpsEndpoint built-in topic.
        """
        return ddspy_get_matched_subscription_data(self._ref, handle, endpoint_constructor, cqos_to_qos)

    def get_liveliness_lost_status(self):
        """Get LIVELINESS_LOST status

        Raises
        ------
        DDSException

        Returns
        -------
        liveness_lost_status:
            The class 'liveness_lost_status' value.
        """
        status = dds_c_t.liveliness_lost_status()
        ret = self._get_liveliness_lost_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the liveliness lost status for {repr(self)}")

    def get_offered_deadline_missed_status(self):
        """Get OFFERED DEADLINE MISSED status

        Raises
        ------
        DDSException

        Returns
        -------
        offered_deadline_missed_status:
            The class 'offered_deadline_missed_status' value.
        """
        status = dds_c_t.offered_deadline_missed_status()
        ret = self._get_offered_deadline_missed_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the offered deadline missed status for {repr(self)}")

    def get_offered_incompatible_qos_status(self):
        """Get OFFERED INCOMPATIBLE QOS status

        Raises
        ------
        DDSException

        Returns
        -------
        offered_incompatible_qos_status:
            The class 'offered_incompatible_qos_status' value.
        """
        status = dds_c_t.offered_incompatible_qos_status()
        ret = self._get_offered_incompatible_qos_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the offered incompatible qos status for {repr(self)}")

    def get_publication_matched_status(self):
        """Get PUBLICATION MATCHED status

        Raises
        ------
        DDSException

        Returns
        -------
        publication_matched_status:
            The class 'publication_matched_status' value.
        """
        status = dds_c_t.publication_matched_status()
        ret = self._get_publication_matched_status(self._ref, ct.byref(status))
        if ret == 0:
            return status
        raise DDSException(ret, f"Occurred when getting the publication matched status for {repr(self)}")

    @c_call("dds_create_writer")
    def _create_writer(self, publisher: dds_c_t.entity, topic: dds_c_t.entity, qos: dds_c_t.qos_p,
                       listener: dds_c_t.listener_p) -> dds_c_t.entity:
        pass

    @c_call("dds_wait_for_acks")
    def _wait_for_acks(self, publisher: dds_c_t.entity, timeout: dds_c_t.duration) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_matched_subscriptions")
    def _get_matched_subscriptions(self, writer: dds_c_t.entity, handle: ct.POINTER(dds_c_t.instance_handle),
                                   size: ct.c_size_t) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_liveliness_lost_status")
    def _get_liveliness_lost_status(self, writer: dds_c_t.entity, status: ct.POINTER(dds_c_t.liveliness_lost_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_offered_deadline_missed_status")
    def _get_offered_deadline_missed_status(self, writer: dds_c_t.entity, status: ct.POINTER(dds_c_t.offered_deadline_missed_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_offered_incompatible_qos_status")
    def _get_offered_incompatible_qos_status(self, writer: dds_c_t.entity, status: ct.POINTER(dds_c_t.offered_incompatible_qos_status)) -> dds_c_t.returnv:
        pass

    @c_call("dds_get_publication_matched_status")
    def _get_publication_matched_status(self, writer: dds_c_t.entity, status: ct.POINTER(dds_c_t.publication_matched_status)) -> dds_c_t.returnv:
        pass



import gc
import json
import time
import sys
import os
import pytest
import shutil
import asyncio
import tempfile
import concurrent
import subprocess

from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.pub import DataWriter
from cyclonedds.sub import DataReader
from cyclonedds.core import Qos, Policy

from support_modules.testtopics import Message


if sys.platform.startswith("win"):
    pytest.skip("DDSLS is unstable on windows", allow_module_level=True)


# Helper functions

def run_ddsls(args, timeout=10):
    tmp_dir = tempfile.mkdtemp()
    process = subprocess.Popen([sys.executable, "-m", "cyclonedds.tools.ddsls"] + args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                                cwd=tmp_dir
                            )
    try:
        stdout, stderr = process.communicate(timeout=timeout)
    except subprocess.TimeoutExpired as e:
        try:
            process.kill()
            shutil.rmtree(tmp_dir)
        except:
            # Observant readers will note that if the process is not killed
            # then we do not delete the temporary dir. This has a good reason:
            # When the process is stuck in an unkillable state and we try to remove
            # the directory, we will enter an infinite loop with permission errors on
            # windows.
            pass
        try:
            print("Ddsls command timeout")
            print("   attempting to pull from stdout:")
            process.stdout.close()
            print(process.stdout.read().decode().replace("\r", ""))
            print("   attempting to pull from stderr")
            process.stderr.close()
            print(process.stderr.read().decode().replace("\r", ""))
        except:
            pass
        raise e

    shutil.rmtree(tmp_dir)

    return {
        "stdout": stdout.decode().replace("\r", ""),
        "stderr": stderr.decode().replace("\r", ""),
        "status": process.returncode
    }


async def run_ddsls_async_watch(args, runtime, runner):
    loop = asyncio.get_event_loop_policy().get_event_loop()
    with concurrent.futures.ThreadPoolExecutor() as pool:
        task = loop.run_in_executor(pool, run_ddsls, ["--watch", "--runtime", str(runtime)] + args)
        await asyncio.sleep(0.3)
        test = await runner()
        return (await task), test


def run_ddsls_watchmode(args, runner, runtime=5):
    loop = asyncio.get_event_loop_policy().get_event_loop()
    result = loop.run_until_complete(run_ddsls_async_watch(args, runtime, runner))
    return result


# Tests

def test_ddsls_participant_runs():
    data = run_ddsls(["-t", "dcpsparticipant"])
    assert data["status"] == 0


def test_ddsls_participant_runs_json():
    data = run_ddsls(["--json", "-t", "dcpsparticipant"])
    assert data["status"] == 0
    json.loads(data["stdout"])


def test_ddsls_publication_runs():
    data = run_ddsls(["-t", "dcpspublication"])
    assert data["status"] == 0


def test_ddsls_publication_runs_json():
    data = run_ddsls(["--json", "-t", "dcpspublication"])
    assert data["status"] == 0
    json.loads(data["stdout"])


def test_ddsls_subscription_runs():
    data = run_ddsls(["-t", "dcpssubscription"])
    assert data["status"] == 0


def test_ddsls_subscription_runs_json():
    data = run_ddsls(["--json", "-t", "dcpssubscription"])
    assert data["status"] == 0
    json.loads(data["stdout"])


def test_ddsls_all_runs():
    data = run_ddsls(["-a"])
    assert data["status"] == 0


def test_ddsls_all_runs_json():
    data = run_ddsls(["--json", "-a"])
    assert data["status"] == 0
    json.loads(data["stdout"])


def test_ddsls_participant_reported():
    dp = DomainParticipant(0)
    time.sleep(0.5)

    data = run_ddsls(["-t", "dcpsparticipant"])

    time.sleep(0.5)

    assert str(dp.guid) in data["stdout"]


def test_ddsls_participant_json_reported():
    dp = DomainParticipant(0)
    time.sleep(0.5)

    data = run_ddsls(["-t", "dcpsparticipant", "--json"])

    time.sleep(0.5)

    assert str(dp.guid) in data["stdout"]


def test_ddsls_participant_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        return dp

    data, dp = run_ddsls_watchmode(["-t", "dcpsparticipant"], test_inner, runtime=1)

    assert str(dp.guid) in data["stdout"]


def test_ddsls_participant_json_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        return dp

    data, dp = run_ddsls_watchmode(["-t", "dcpsparticipant", "--json"], test_inner, runtime=1)

    assert str(dp.guid) in data["stdout"]


def test_ddsls_subscription_reported():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dr = DataReader(dp, tp)
    time.sleep(1)

    data = run_ddsls(["-t", "dcpssubscription"])

    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]
    assert str(dr.get_qos()) in data["stdout"]


def test_ddsls_subscription_json_reported():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dr = DataReader(dp, tp)
    time.sleep(1)

    data = run_ddsls(["--json", "-t", "dcpssubscription"])
    json_data = json.loads(data["stdout"])

    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]

    reader_check = False
    for sample in json_data:
        for val in sample["value"]:
            if val["key"] == str(dr.guid):
                assert (dr.get_qos()).asdict() == val["qos"]
                reader_check = True
    assert reader_check


def test_ddsls_subscription_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dr = DataReader(dp, tp)
        return dp, tp, dr

    data, (dp, tp, dr) = run_ddsls_watchmode(["-t", "dcpssubscription"], test_inner, runtime=1)

    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]
    assert str(dr.get_qos()) in data["stdout"]


def test_ddsls_subscription_json_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dr = DataReader(dp, tp)
        return dp, tp, dr

    data, (dp, tp, dr) = run_ddsls_watchmode(["-t", "dcpssubscription", "--json"], test_inner, runtime=1)
    json_data = json.loads(data["stdout"])

    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]

    reader_check = False
    for sample in json_data:
        for val in sample["value"]:
            if val["key"] == str(dr.guid):
                assert (dr.get_qos()).asdict() == val["qos"]
                reader_check = True
    assert reader_check



def test_ddsls_publication_reported():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dw = DataWriter(dp, tp)
    time.sleep(1)

    data = run_ddsls(["-t", "dcpspublication"])

    assert str(dw.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]
    assert str(dw.get_qos()) in data["stdout"]


def test_ddsls_publication_json_reported():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dw = DataWriter(dp, tp)
    time.sleep(1)

    data = run_ddsls(["--json", "-t", "dcpspublication"])
    json_data = json.loads(data["stdout"])

    assert str(dw.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]

    writer_check = False
    for sample in json_data:
        for val in sample["value"]:
            if val["key"] == str(dw.guid):
                assert (dw.get_qos()).asdict() == val["qos"]
                writer_check = True
    assert writer_check


def test_ddsls_publication_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp)
        return dp, tp, dw

    data, (dp, tp, dw) = run_ddsls_watchmode(["-t", "dcpspublication"], test_inner, runtime=1)

    assert str(dw.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]
    assert str(dw.get_qos()) in data["stdout"]


def test_ddsls_publication_json_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp)
        return dp, tp, dw

    data, (dp, tp, dw) = run_ddsls_watchmode(["-t", "dcpspublication", "--json"], test_inner, runtime=1)
    json_data = json.loads(data["stdout"])

    assert str(dw.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]

    writer_check = False
    for sample in json_data:
        for val in sample["value"]:
            if val["key"] == str(dw.guid):
                assert (dw.get_qos()).asdict() == val["qos"]
                writer_check = True
    assert writer_check


def test_ddsls_all_entities_reported():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dw = DataWriter(dp, tp)
    dr = DataReader(dp, tp)
    time.sleep(1)

    data = run_ddsls(["-a"])

    assert str(dw.guid) in data["stdout"]
    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]
    assert str(dw.get_qos()) in data["stdout"]
    assert str(dr.get_qos()) in data["stdout"]


def test_ddsls_all_entities_json_reported():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dw = DataWriter(dp, tp)
    dr = DataReader(dp, tp)
    time.sleep(1)

    data = run_ddsls(["--json", "-a"])
    json_data = json.loads(data["stdout"])

    assert str(dw.guid) in data["stdout"]
    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]

    writer_check, reader_check = False, False
    for sample in json_data:
        for val in sample["value"]:
            if val["key"] == str(dw.guid):
                assert (dw.get_qos()).asdict() == val["qos"]
                writer_check = True
            if val["key"] == str(dr.guid):
                assert (dr.get_qos()).asdict() == val["qos"]
                reader_check = True
    assert reader_check and writer_check


def test_ddsls_all_entities_watch_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp)
        dr = DataReader(dp, tp)
        return dp, tp, dw, dr

    data, (dp, tp, dw, dr) = run_ddsls_watchmode(["-a"], test_inner, runtime=2)

    assert str(dw.guid) in data["stdout"]
    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]
    assert str(dw.get_qos()) in data["stdout"]
    assert str(dr.get_qos()) in data["stdout"]


def test_ddsls_all_entities_watch_json_reported():
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp)
        dr = DataReader(dp, tp)
        return dp, tp, dw, dr

    data, (dp, tp, dw, dr) = run_ddsls_watchmode(["-a", "--json"], test_inner, runtime=2)
    json_data = json.loads(data["stdout"])

    assert str(dw.guid) in data["stdout"]
    assert str(dr.guid) in data["stdout"]
    assert str(dp.guid) in data["stdout"]
    assert tp.name in data["stdout"]
    assert tp.typename in data["stdout"]

    writer_check, reader_check = False, False
    for sample in json_data:
        for val in sample["value"]:
            if val["key"] == str(dw.guid):
                assert (dw.get_qos()).asdict() == val["qos"]
                writer_check = True
            if val["key"] == str(dr.guid):
                assert (dr.get_qos()).asdict() == val["qos"]
                reader_check = True
    assert reader_check and writer_check


def test_ddsls_write_to_file(tmp_path):
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageTopic", Message)
    dw = DataWriter(dp, tp)
    dr = DataReader(dp, tp)

    time.sleep(0.5)

    run_ddsls(["--json", "-a", "--filename", str(tmp_path / "test.json")])

    time.sleep(0.5)

    with open(tmp_path / "test.json") as f:
        data = json.load(f)

    assert str(dw.guid) in data["PUBLICATION"]["New"]
    assert str(dr.guid) in data["SUBSCRIPTION"]["New"]
    assert str(dp.guid) in data["PARTICIPANT"]["New"]
    assert tp.name in data["PUBLICATION"]["New"][str(dw.guid)]["topic_name"]
    assert tp.typename in data["PUBLICATION"]["New"][str(dw.guid)]["type_name"]
    assert tp.name in data["SUBSCRIPTION"]["New"][str(dr.guid)]["topic_name"]
    assert tp.typename in data["SUBSCRIPTION"]["New"][str(dr.guid)]["type_name"]


def test_ddsls_write_disposed_data_to_file(tmp_path):
    async def test_inner():
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp)
        dr = DataReader(dp, tp)

        disposed_data = {
            "dp.guid": str(dp.guid),
            "tp.name": tp.name,
            "tp.typename": tp.typename,
            "dw.guid": str(dw.guid),
            "dr.guid": str(dr.guid)
        }
        await asyncio.sleep(1)

        del dp, tp, dw, dr
        gc.collect()
        await asyncio.sleep(1)
        return disposed_data

    procdata, disposed_data = run_ddsls_watchmode(["--json", "-a", "--filename", str(tmp_path / "test_disposed.json")],
                                           test_inner, runtime=2)

    assert procdata['status'] == 0

    with open(tmp_path / "test_disposed.json") as f:
        data = json.load(f)

    dw_guid = disposed_data["dw.guid"]
    dr_guid = disposed_data["dr.guid"]

    assert disposed_data["dp.guid"] not in data["PARTICIPANT"]["New"]
    assert disposed_data["dp.guid"] in data["PARTICIPANT"]["Disposed"]
    assert dw_guid not in data["PUBLICATION"]["New"]
    assert dw_guid in data["PUBLICATION"]["Disposed"]
    assert dr_guid not in data["SUBSCRIPTION"]["New"]
    assert dr_guid in data["SUBSCRIPTION"]["Disposed"]
    assert disposed_data["tp.name"] == data["PUBLICATION"]["Disposed"][dw_guid]["topic_name"]
    assert disposed_data["tp.name"] == data["SUBSCRIPTION"]["Disposed"][dr_guid]["topic_name"]
    assert disposed_data["tp.typename"] == data["PUBLICATION"]["Disposed"][dw_guid]["type_name"]
    assert disposed_data["tp.typename"] == data["SUBSCRIPTION"]["Disposed"][dr_guid]["type_name"]


def test_ddsls_domain_id():
    dp = DomainParticipant(0)
    dp1 = DomainParticipant(33)

    time.sleep(0.5)

    data = run_ddsls(["--id", "33", "-t", "dcpsparticipant"])

    time.sleep(0.5)

    assert str(dp.guid) not in data["stdout"]
    assert str(dp1.guid) in data["stdout"]


def test_ddsls_qos_change():
    async def test_inner():
        qos = Qos(Policy.OwnershipStrength(10),
                  Policy.Userdata("Old".encode()))
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp, qos=qos)
        await asyncio.sleep(0.5)

        old_qos = dw.get_qos()
        await asyncio.sleep(0.5)

        new_qos = Qos(Policy.OwnershipStrength(20),
                      Policy.Userdata("New".encode()))
        dw.set_qos(new_qos)

        await asyncio.sleep(0.5)

        return dp, tp, dw, old_qos, new_qos

    data, (dp, tp, dw, old_qos, new_qos) = run_ddsls_watchmode(["-a"], test_inner, runtime=5)

    assert str(old_qos) in data["stdout"]
    for q in new_qos:
        assert str(q) in data["stdout"]
        assert f"{str(old_qos[q])} -> {str(q)}" in data["stdout"]


def test_ddsls_qos_change_in_verbose():
    async def test_inner():
        qos = Qos(Policy.OwnershipStrength(10),
                  Policy.Userdata("Old".encode()))
        dp = DomainParticipant(0)
        tp = Topic(dp, "MessageTopic", Message)
        dw = DataWriter(dp, tp, qos=qos)
        await asyncio.sleep(0.5)

        old_qos = dw.get_qos()
        await asyncio.sleep(0.5)

        new_qos = Qos(Policy.OwnershipStrength(20),
                      Policy.Userdata("New".encode()))
        dw.set_qos(new_qos)

        await asyncio.sleep(0.5)

        return dp, tp, dw, old_qos, new_qos

    data, (dp, tp, dw, old_qos, new_qos) = run_ddsls_watchmode(["-a", "--verbose"], test_inner, runtime=5)

    assert str(old_qos) in data["stdout"]
    for q in new_qos:
        assert str(q) in data["stdout"]
        assert f"{str(old_qos[q])} -> {str(q)}" in data["stdout"]
    assert str(dw.get_qos()) in data["stdout"]


# test error messages


def test_ddsls_file_open_error(tmp_path):
    data = run_ddsls(["--json", "-a", "--filename", f"{tmp_path}/this/path/denfinitely/doesnot/exist/ever"])
    assert f"Exception: Could not open file {tmp_path}/this/path/denfinitely/doesnot/exist/ever" in data["stderr"]



import pytest

from cyclonedds.core import Entity, DDSException, WaitSet, ReadCondition, ViewState, InstanceState, SampleState
from cyclonedds.util import duration, isgoodentity

from support_modules.testtopics import Message


def test_waitset_initialize(common_setup):
    ws = WaitSet(common_setup.dp)

    assert isgoodentity(ws)


def test_waitset_attachment(common_setup):
    ws = WaitSet(common_setup.dp)
    rc = ReadCondition(common_setup.dr, ViewState.Any | InstanceState.Any | SampleState.Any)

    ws.attach(rc)
    assert ws.is_attached(rc)

    ws.detach(rc)
    assert not ws.is_attached(rc)


def test_waitset_illigal_op(common_setup):
    ws = WaitSet(common_setup.dp)

    with pytest.raises(DDSException) as exc:
        ws.attach(Entity(101))

    assert exc.value.code == DDSException.DDS_RETCODE_BAD_PARAMETER


def test_waitset_wait(common_setup):
    ws = WaitSet(common_setup.dp)

    rc1 = ReadCondition(common_setup.dr, ViewState.Any | InstanceState.Any | SampleState.Any)
    ws.attach(rc1)

    assert ws.wait(duration(milliseconds=5)) == 0

    common_setup.dw.write(Message(message="Hi!"))

    assert ws.wait(duration(seconds=1)) == 1

    rc2 = ReadCondition(common_setup.dr, ViewState.Any | InstanceState.Any | SampleState.NotRead)
    ws.attach(rc2)

    assert ws.wait(duration(seconds=1)) == 2


from cyclonedds.idl import IdlUnion, types


class A(IdlUnion, discriminator=types.uint8):
    f1: types.case[1, types.uint8]
    f2: types.default[types.uint16]


def test_union_not_equal():
    # Even though these both map to the default case
    # The very "thoughtful" XTypes API maps these to type kinds without case
    # Which is very nice...
    assert A(discriminator=3, value=1) != A(discriminator=2, value=1)




from .discovery.type_discoverables import TypeDiscoveryData, DiscoveredType
from typing import Optional
from rich.syntax import Syntax
from rich.prompt import Prompt
from rich.console import Console
from cyclonedds.qos import Qos, Policy
from .qosformat import rich_qos
import random


def select_type(
    console: Console, data: Optional[TypeDiscoveryData], pick_random: bool
) -> Optional[DiscoveredType]:
    types = list(data.types.values())
    if data is None or len(types) == 0:
        console.print(
            "[bold red] :police_car_light: No types could be discovered over XTypes, no dynamism possible[/]"
        )
        return
    elif len(types) == 1:
        return types[0]
    elif pick_random:
        return random.choice(types)

    console.print(
        "[bold orange] :warning: Multiple type definitions exist, please pick one"
    )

    for i, dtype in enumerate(types):
        console.print(
            f"Type {i}, As defined in participant(s) [magenta]"
            + ", ".join(f"{p}" for p in dtype.participants)
        )
        console.print(Syntax(dtype.code, "omg-idl"))
        console.print()

    index = Prompt.ask(
        "Please pick a type",
        choices=[str(i) for i in range(len(types))],
    )
    index = int(index)
    return types[index]


def _filter_non_override(base_qos: Qos, override_qos: Qos) -> Qos:
    return Qos(
        *[p for p in override_qos.policies if p not in base_qos or base_qos[p] != p]
    )


def select_qos(
    console: Console,
    data: Optional[TypeDiscoveryData],
    for_writer: bool,
    pick_random: bool,
) -> Qos:
    topic_qos_shared, topic_qos_mix = data.split_qos(data.topic_qosses)

    if (for_writer and data.writer_qosses) or not data.reader_qosses:
        endpoint_qosses = data.writer_qosses
    else:
        endpoint_qosses = data.reader_qosses

    endpoint_qos_shared, endpoint_qos_mix = data.split_qos(endpoint_qosses)

    if len(topic_qos_mix) <= 1:
        topic_qos = topic_qos_shared
    elif pick_random:
        topic_qos = topic_qos_shared + random.choice(topic_qos_mix)
    else:
        console.print(
            "[bold orange] :warning: Multiple Topic QoS configurations, please pick one"
        )

        if topic_qos_shared:
            console.print("[bold green]Common Topic QoS")
            console.print(rich_qos(topic_qos_shared))

        console.print("[bold green]Variants")

        for i, qos in enumerate(topic_qos_mix):
            console.print(f"[bold bright_cyan]Variant {i}")
            console.print(rich_qos(qos))

        index = Prompt.ask(
            "Please pick a QoS variant",
            choices=[str(i) for i in range(len(topic_qos_mix))],
        )
        index = int(index)
        topic_qos = topic_qos_shared + topic_qos_mix[index]

    if len(endpoint_qos_mix) <= 1:
        return topic_qos, endpoint_qos_shared
    elif pick_random:
        return topic_qos, endpoint_qos_shared + random.choice(endpoint_qos_mix)

    console.print(
        "[bold orange] :warning: Multiple Endpoint QoS configurations, please pick one"
    )

    shared_qos = topic_qos + endpoint_qos_shared

    if shared_qos:
        console.print("[bold green]Common Endpoint QoS")
        console.print(rich_qos(shared_qos))

    console.print("[bold green]Variants")

    for i, qos in enumerate(endpoint_qos_mix):
        console.print(f"[bold bright_cyan]Variant {i}")
        console.print(rich_qos(_filter_non_override(shared_qos, qos)))

    index = Prompt.ask(
        "Please pick a QoS variant",
        choices=[str(i) for i in range(len(endpoint_qos_mix))],
    )
    index = int(index)

    return topic_qos, shared_qos + endpoint_qos_mix[index]



import pytest

from cyclonedds.core import Entity, ReadCondition, SampleState, InstanceState, ViewState
from cyclonedds.util import isgoodentity
from cyclonedds.topic import Topic
from cyclonedds.sub import DataReader
from cyclonedds.pub import DataWriter

from support_modules.testtopics import Message, MessageKeyed


def test_readcondition_init(common_setup):
    rc = ReadCondition(common_setup.dr, SampleState.Any | InstanceState.Any | ViewState.Any)
    assert isgoodentity(rc)


def test_readcondition_get_mask(common_setup):
    mask = SampleState.Any | InstanceState.Any | ViewState.Any
    rc = ReadCondition(common_setup.dr, mask)

    assert rc.mask == rc.get_mask() == mask

    mask = SampleState.NotRead | InstanceState.NotAliveNoWriters | ViewState.Old

    assert rc.mask == rc.get_mask() != mask

    rc = ReadCondition(common_setup.dr, mask)

    assert rc.mask == rc.get_mask() == mask


def test_readcondition_get_reader(common_setup):
    rc = ReadCondition(common_setup.dr, SampleState.Any | InstanceState.Any | ViewState.Any)
    assert rc.get_datareader() == common_setup.dr


def test_readcondition_read(common_setup):
    tp = Topic(common_setup.dp, "hi_sayer", MessageKeyed)
    dr = DataReader(common_setup.dp, tp)
    dw = DataWriter(common_setup.dp, tp)
    rc = ReadCondition(dr, SampleState.Any | ViewState.Any | InstanceState.NotAliveDisposed)

    assert not rc.triggered

    messages = [MessageKeyed(user_id=i, message=f"Hi {i}!") for i in range(5)]
    for m in messages:
        dw.write(m)

    received = dr.read(N=5)

    assert messages == received

    dw.dispose(messages[1])
    assert rc.triggered

    received = dr.read(condition=rc)

    assert len(received) == 1 and received[0] == messages[1]


def test_condition_cleanup(common_setup):
    reader = DataReader(common_setup.dp, common_setup.tp)
    condition = ReadCondition(reader, 1234)

    del reader
    import gc
    gc.collect()

    del condition


"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from typing import Optional, cast, Any, ClassVar, Mapping, Dict, Tuple, TYPE_CHECKING
from collections import deque
from enum import EnumMeta, Enum
from inspect import isclass
from struct import unpack
from hashlib import md5
import threading

from ._support import Buffer, Endianness, CdrKeyVmNamedJumpOp, KeyScanner, KeyScanResult, SerializeKind, DeserializeKind, DataTypeProperties
from ._type_helper import get_origin, get_args, Annotated
from ._type_normalize import get_idl_annotations, get_idl_field_annotations, get_extended_type_hints
from ._machinery import Machine

from . import types


if TYPE_CHECKING:
    from ._typesupport.DDS.XTypes import TypeMapping, TypeInformation, TypeIdentifier
    from ._builder import XCDRSupported
    from cyclonedds.idl import IdlEnum


class IDLNamespaceScope:
    current = None
    stack = deque()

    @classmethod
    def enter(cls, scope):
        if cls.current is not None:
            cls.stack.append(cls.current)
        cls.current = scope

    @classmethod
    def exit(cls):
        if cls.stack:
            cls.current = cls.stack.pop()
        else:
            cls.current = None


class IDL:
    def __init__(self, datatype):
        self._populated: bool = False
        self._lock = threading.RLock()
        self._populating: bool = False
        self.datatype: type = datatype
        self.keyless: bool = None
        self.v1_machine: Machine = None
        self.v2_machine: Machine = None
        self.v1_key_max_size: int = None
        self.v2_key_max_size: int = None
        self.supported_versions: int = None
        self.default_version: int = None
        self.data_type_props: int = None

        self.idl_transformed_typename: str = self.datatype.__idl_typename__.replace(".", "::")
        self.re_entrancy_protection: bool = False
        self._xt_data: Tuple[TypeInformation, TypeMapping] = (None, None)
        self._xt_bytedata: Tuple[Optional[bytes], Optional[bytes]] = (None, None)
        self.member_ids: Dict[str, int] = None

    def populate_locked(self):
        if not self._populating:
            self._populating = True
            annotations = get_idl_annotations(self.datatype)
            field_annotations = get_idl_field_annotations(self.datatype)

            a = annotations.get('extensibility', 'final')
            if a == 'appendable':
                self.xcdrv1_head = 0x00
                self.xcdrv2_head = 0x08
            elif a == 'mutable':
                self.xcdrv1_head = 0x02
                self.xcdrv2_head = 0x0a
            else:
                self.xcdrv1_head = 0x00
                self.xcdrv2_head = 0x06

            if self.member_ids is None:
                ids = {}
                is_hash_id = annotations.get("autoid", "sequential") == "hash"
                idc = 0

                for name, _ in get_extended_type_hints(self.datatype).items():
                    f_annot = field_annotations.get(name, {})

                    if "id" in f_annot:
                        mid = f_annot["id"]
                    elif "hash_id" in f_annot or is_hash_id:
                        # compute 4 byte hash, interpret as little endian 32 bit integer and zero out top four bits
                        mid = unpack("<I", md5(f_annot.get("hash_id", "") or name.encode()).digest()[:4])[0] & 0x0FFFFFFF
                    else:
                        mid = idc

                    idc = mid + 1
                    ids[name] = mid

                self.member_ids = ids

            from ._builder import Builder
            self.v1_machine, self.v2_machine, self.data_type_props, self.supported_versions, self.default_version = Builder.build_machines(self.datatype)
            self.keyless = (self.data_type_props & DataTypeProperties.CONTAINS_KEY) == 0

            self.v1_keyresult: KeyScanner = self.v1_machine.key_scan()
            if self.v1_keyresult.rtype != KeyScanResult.PossiblyInfinite and self.v1_keyresult.size <= 16:
                self.v1_key_max_size = self.v1_keyresult.size
            else:
                self.v1_key_max_size = 17  # or bigger ;)

            self.v2_keyresult: KeyScanner = self.v2_machine.key_scan()
            if self.v2_keyresult.rtype != KeyScanResult.PossiblyInfinite and self.v2_keyresult.size <= 16:
                self.v2_key_max_size = self.v2_keyresult.size
            else:
                self.v2_key_max_size = 17  # or bigger ;)

    def populate(self):
        with self._lock:
            self.populate_locked()
        # hopefully the memory order guarantees of Python are strong enough to make it
        # impossible for another thread to observe a partially populated self
        self._populated = True

    def serialize(self, object, use_version_2: bool = None, buffer=None, endianness : Endianness = None, serialize_kind: SerializeKind = SerializeKind.DataSample, prepend_header: bool = True) -> bytes:
        if not self._populated:
            self.populate()

        if use_version_2 is None:
            use_version_2 = (self.default_version == 2)

        ibuffer = buffer or Buffer()
        ibuffer.seek(0)
        ibuffer.zero_out()
        ibuffer.set_align_offset(0)
        ibuffer.set_endianness(endianness or Endianness.native())
        ibuffer._align_max = 4 if use_version_2 else 8

        if prepend_header:
            enc = ((0 if ibuffer.endianness == Endianness.Big else 1) |
                   (self.xcdrv2_head if use_version_2 else self.xcdrv1_head))
            ibuffer.write('b', 1, 0)
            ibuffer.write('b', 1, enc)
            ibuffer.write('b', 1, 0)
            ibuffer.write('b', 1, 0)
            ibuffer.set_align_offset(4)

        if use_version_2:
            self.v2_machine.serialize(ibuffer, object, serialize_kind)
        else:
            self.v1_machine.serialize(ibuffer, object, serialize_kind)

        return ibuffer.asbytes()

    def deserialize(self, data, has_header=True, use_version_2: bool = None, deserialize_kind: DeserializeKind = DeserializeKind.DataSample) -> object:
        if not self._populated:
            self.populate()

        if has_header and use_version_2 is not None:
            raise Exception("Considered programmer error to set a version of xcdr to use if a header is present in the data.")
        if use_version_2 is None:
            use_version_2 = (self.default_version == 2)

        buffer = Buffer(data, align_offset=4 if has_header else 0) if not isinstance(data, Buffer) else data

        if has_header and buffer.tell() == 0:
            buffer.read('b', 1)
            v = buffer.read('b', 1)
            if (v & 1) > 0:
                buffer.set_endianness(Endianness.Little)
            else:
                buffer.set_endianness(Endianness.Big)
            buffer.read('b', 1)
            buffer.read('b', 1)
            if v > 3:
                buffer._align_max = 4
                machine = self.v2_machine
            else:
                buffer._align_max = 8
                machine = self.v1_machine
        else:
            if use_version_2:
                buffer._align_max = 4
                machine = self.v2_machine
            else:
                buffer._align_max = 8
                machine = self.v1_machine

        return machine.deserialize(buffer, deserialize_kind=deserialize_kind)

    def deserialize_key(self, data, has_header=True, use_version_2: bool = None) -> object:
        return self.deserialize(data, has_header, use_version_2, DeserializeKind.KeySample)

    def serialize_key(self, object, use_version_2: bool = None, endianness: Endianness = Endianness.Little) -> bytes:
        return self.serialize(object, use_version_2, None, endianness, SerializeKind.KeyDefinitionOrder, False)

    def serialize_key_normalized(self, object, use_version_2: bool = None, endianness: Endianness = Endianness.Little) -> bytes:
        return self.serialize(object, use_version_2, None, endianness, SerializeKind.KeyNormalized, False)

    def cdr_key_machine(self, skip: bool = False, use_version_2: bool = None):
        if self.re_entrancy_protection:
            # If we get here then there is a recursion in the type
            # We will need to use a jump instruction
            return [CdrKeyVmNamedJumpOp(jump_to=self.datatype.__idl_typename__)]

        if not self._populated:
            self.populate()

        if use_version_2 is None:
            use_version_2 = (self.default_version == 2)

        self.re_entrancy_protection = True

        if use_version_2:
            ops = self.v2_machine.cdr_key_machine_op(skip)
        else:
            ops = self.v1_machine.cdr_key_machine_op(skip)

        self.re_entrancy_protection = False

        return ops

    def key_scan(self, use_version_2: bool = None):
        if self.re_entrancy_protection:
            # If we get here then there is a recursion in the type
            # This always means the keysize can be infinite
            return KeyScanner.infinity()

        if not self._populated:
            self.populate()

        if use_version_2 is None:
            use_version_2 = (self.default_version == 2)

        self.re_entrancy_protection = True

        if use_version_2:
            scan = self.v2_machine.key_scan()
        else:
            scan = self.v1_machine.key_scan()

        self.re_entrancy_protection = False

        return scan

    def get_member_id(self, member: str) -> int:
        return self.member_ids.get(member, -1) if self.member_ids else -1

    def fill_type_data(self) -> None:
        if not self._populated:
            self.populate()

        if self._xt_data[0] is None:
            from ._xt_builder import XTBuilder
            self._xt_data = XTBuilder.process_type(self.datatype)
            self._xt_bytedata = (
                self._xt_data[0].serialize(endianness=Endianness.Little, use_version_2=True)[4:],
                self._xt_data[1].serialize(endianness=Endianness.Little, use_version_2=True)[4:]
            )

    def get_type_info(self) -> 'TypeInformation':
        if self._xt_data[0] is None:
            self.fill_type_data()
        return self._xt_data[0]

    def get_type_mapping(self) -> 'TypeMapping':
        if self._xt_data[0] is None:
            self.fill_type_data()
        return self._xt_data[1]

    def get_type_id(self) -> 'TypeIdentifier':
        if self._xt_data[0] is None:
            self.fill_type_data()
        return self._xt_data[0].complete.typeid_with_size.type_id


class IdlMeta(type):
    __idl__: ClassVar['IDL']
    __idl_typename__: ClassVar[str]
    __idl_annotations__: ClassVar[Dict[str, Any]]
    __idl_field_annotations__: ClassVar[Dict[str, Dict[str, Any]]]

    @classmethod
    def __prepare__(metacls, __name: str, __bases: Tuple[type, ...], **kwds: Any) -> Mapping[str, Any]:
        typename = None
        if "typename" in kwds:
            typename = kwds["typename"]
            del kwds["typename"]

        namespace = super().__prepare__(__name, __bases, **kwds)
        namespace["__annotations__"] = {}
        namespace["__idl_annotations__"] = {}
        namespace["__idl_field_annotations__"] = {}
        if typename:
            namespace["__idl_typename__"] = typename

        for base in __bases:
            if isinstance(base, IdlMeta):
                namespace["__annotations__"].update(base.__annotations__)
                namespace["__idl_annotations__"].update(base.__idl_annotations__)
                namespace["__idl_field_annotations__"].update(base.__idl_field_annotations__)

        IDLNamespaceScope.enter(namespace)
        return namespace

    def __new__(metacls, name, bases, namespace, **kwds):
        IDLNamespaceScope.exit()
        new_cls = super().__new__(metacls, name, bases, dict(**namespace))

        if "__idl_typename__" not in namespace:
            new_cls.__idl_typename__ = name

        new_cls.__idl__ = IDL(new_cls)
        return new_cls

    def __repr__(cls):
        # Note, this is the _class_ repr
        if cls.__name__ == "IdlStruct":
            return "IdlStruct"
        return f"{cls.__name__}(IdlStruct, idl_typename='{cls.__idl_typename__}')"


def _union_default_finder(_type, cases):
    if isclass(_type) and issubclass(_type, Enum):
        # non-optimal for sure, but should be used rarely
        used = set(e.value for e in _type)
        i = 0
        while i in used:
            i += 1
        # TODO: bit bound check?
        return i

    if _type == bool:
        if True not in cases:
            return True
        elif False not in cases:
            return False
        raise TypeError("No space in discriminated union for default value.")

    val, inc, end = {
        types.int8: (-1, -1, -128),
        types.int16: (-1, -1, -32768),
        types.int32: (-1, -1, -2147483648),
        types.int64: (-1, -1, -9223372036854775808),
        int: (-1, -1, -9223372036854775808),
        types.byte: (0, 1, 255),
        types.uint8: (0, 1, 255),
        types.uint16: (0, 1, 65535),
        types.uint32: (0, 1, 4294967295),
        types.uint64: (0, 1, 18446744073709551615),
        types.char: (0, 1, 127)
    }.get(_type, (None, None, None))

    if val is None:
        raise TypeError(f"{_type} is not a valid discriminator type for an IdlUnion.")

    while True:
        if val not in cases:
            if _type == types.char:
                return chr(val)
            return val
        if val == end:
            raise TypeError("No space in discriminated union for default value.")
        val += inc


class IdlUnionMeta(IdlMeta):
    __idl_discriminator__: Any
    __idl_discriminator_is_key__: bool
    __idl_cases__: Dict[Any, Any]
    __idl_default__: Optional[Tuple[Any, Any]]

    @classmethod
    def __prepare__(metacls, __name: str, __bases: Tuple[type, ...], **kwds: Any) -> Mapping[str, Any]:
        if not len(__bases):
            return super().__prepare__(__name, __bases, **kwds)

        discriminator_is_key = False

        if "discriminator" not in kwds:
            raise TypeError("Union class needs a 'discriminator=type'")

        if "discriminator_is_key" in kwds:
            discriminator_is_key = kwds["discriminator_is_key"]
            del kwds["discriminator_is_key"]

        discriminator = kwds["discriminator"]
        del kwds["discriminator"]

        namespace = cast(Dict[str, Any], super().__prepare__(__name, __bases, **kwds))
        namespace["__idl_discriminator__"] = discriminator
        namespace["__idl_discriminator_is_key__"] = discriminator_is_key
        namespace["__annotations__"] = {}
        namespace["__idl_annotations__"] = {}
        namespace["__idl_field_annotations__"] = {}
        IDLNamespaceScope.enter(namespace)
        return namespace

    def __new__(metacls, name, bases, namespace, **kwds):
        IDLNamespaceScope.exit()
        new_cls = super().__new__(metacls, name, bases, dict(**namespace))
        if not len(bases):
            return new_cls

        cases = {}
        default = None
        names = set()

        # Use RAW annotations here because the type strings can maybe not be resolved yet
        for name, _type in new_cls.__annotations__.items():
            if get_origin(_type) != Annotated and len(get_args(_type)) != 2:
                raise TypeError(f"Fields of a union need to be case or default, '{name}: {_type}' is not.")
            if name in ['value', 'discriminator']:
                raise TypeError(f"Field name '{name}' is disallowed in a union, perhaps '_{name}' was meant?")

            _type = get_args(_type)[1]
            if isinstance(_type, types.case):
                for label in _type.labels:
                    if label in cases:
                        raise TypeError(f"Discriminator values must uniquely define a case, "
                                        f"but the case {label} occurred multiple times.")
                    cases[label] = (name, _type.subtype)
                    names.add(name)
            elif isinstance(_type, types.default):
                if default:
                    raise TypeError("A discriminated union can only have one default.")
                default = (name, _type.subtype)
                names.add(name)
            else:
                raise TypeError(f"Fields of a union need to be case or default, '{name}: {_type}' is not.")

        new_cls.__idl_cases__ = cases
        new_cls.__idl_default__ = default
        new_cls.__idl_names__ = names
        new_cls.__idl_discriminator__ = namespace['__idl_discriminator__']
        new_cls.__idl_default_discriminator__ = _union_default_finder(namespace['__idl_discriminator__'], cases)

        return new_cls

    def __repr__(cls):
        # Note, this is the _class_ repr
        if cls.__name__ == "IdlUnion":
            return "IdlUnion"
        return f"{cls.__name__}(IdlUnion, idl_typename='{cls.__idl_typename__}')"


class IdlBitmaskMeta(type):
    __idl_typename__: ClassVar[str]
    __idl_annotations__: ClassVar[Dict[str, Any]]
    __idl_field_annotations__: ClassVar[Dict[str, Dict[str, Any]]]
    __idl_bits__: ClassVar[Dict[int, str]]
    __idl_positions__: ClassVar[Dict[str, int]]
    __idl_highest_position__: ClassVar[int]

    @classmethod
    def __prepare__(metacls, __name: str, __bases: Tuple[type, ...], **kwds: Any) -> Mapping[str, Any]:
        typename = None
        if "typename" in kwds:
            typename = kwds["typename"]
            del kwds["typename"]

        namespace = super().__prepare__(__name, __bases, **kwds)
        namespace["__annotations__"] = {}
        namespace["__idl_annotations__"] = {}
        namespace["__idl_field_annotations__"] = {}
        IDLNamespaceScope.enter(namespace)

        if typename:
            namespace["__idl_typename__"] = typename

        return namespace

    def __new__(metacls, name, bases, namespace, **kwds):
        IDLNamespaceScope.exit()
        new_cls = super().__new__(metacls, name, bases, dict(**namespace))

        if "__idl_typename__" not in namespace:
            new_cls.__idl_typename__ = name

        idl_bits = {}
        idl_positions = {}
        position = 0
        idl_highest_position = None

        if "bit_bound" not in new_cls.__idl_annotations__:
            new_cls.__idl_annotations__["bit_bound"] = 32

        for name, type in new_cls.__annotations__.items():
            if get_origin(type) == ClassVar:
                continue

            if type != bool:
                raise TypeError("Bitmask type contains non-bools.")

            if name in new_cls.__idl_field_annotations__:
                if 'position' in new_cls.__idl_field_annotations__[name]:
                    position = new_cls.__idl_field_annotations__[name]["position"]

            if position < 0 or position >= new_cls.__idl_annotations__["bit_bound"]:
                raise TypeError(
                    f"Position {position} for {name} is outside of the bit boundary"
                    f"[0, {new_cls.__idl_annotations__['bit_bound']})."
                )

            idl_bit = 2 ** position
            if idl_bit in idl_bits:
                raise TypeError(f"Duplicate bit for position {position}.")

            idl_bits[idl_bit] = name
            idl_positions[name] = position
            idl_highest_position = max(idl_highest_position, position) if idl_highest_position is not None else position
            position += 1

        new_cls.__idl_bits__ = idl_bits
        new_cls.__idl_highest_position__ = idl_highest_position
        new_cls.__idl_positions__ = idl_positions

        return new_cls

    def __repr__(cls):
        # Note, this is the _class_ repr
        if cls.__name__ == "IdlBitmask":
            return "IdlBitmask"
        return f"{cls.__name__}(IdlBitmask, idl_typename='{cls.__idl_typename__}')"


class IdlEnumMeta(EnumMeta):
    __idl_typename__: ClassVar[str]
    __idl_annotations__: ClassVar[Dict[str, Any]]
    __idl_field_annotations__: ClassVar[Dict[str, Dict[str, Any]]]
    __idl_enum_default_value__: ClassVar[Optional['IdlEnum']]

    @classmethod
    def __prepare__(metacls, __name: str, __bases: Tuple[type, ...], **kwds: Any) -> Mapping[str, Any]:
        typename = None
        if "typename" in kwds:
            typename = kwds["typename"]
            del kwds["typename"]

        default_ = None
        if "default" in kwds:
            default_ = kwds["default"]
            del kwds["default"]

        namespace: Dict[str, Any] = super().__prepare__(__name, __bases, **kwds)

        if typename:
            namespace["__idl_typename__"] = typename

        namespace["__idl_annotations__"] = {}
        namespace["__idl_field_annotations__"] = {}
        namespace["__idl_enum_default_value__"] = default_
        IDLNamespaceScope.enter(namespace)

        return namespace

    def __new__(metacls, name, bases, namespace, **kwds):
        IDLNamespaceScope.exit()
        new_cls = super().__new__(metacls, name, bases, namespace)

        if "__idl_typename__" not in namespace:
            new_cls.__idl_typename__ = name
        else:
            new_cls.__idl_typename__ = namespace["__idl_typename__"]

        if namespace.get("__idl_enum_default_value__"):
            new_cls.__idl_enum_default_value__ = new_cls[namespace.get("__idl_enum_default_value__")].value
        else:
            values = [entry.value for entry in new_cls]
            if len(values) > 0:
                new_cls.__idl_enum_default_value__ = values[0]

        return new_cls

    _default = object()

    def __call__(cls, value=_default, *args, **kwargs):
        if value is IdlEnumMeta._default:
            return cls.__idl_enum_default_value__
        return super().__call__(value, *args, **kwargs)

    def __repr__(cls):
        # Note, this is the _class_ repr
        if cls.__name__ == "IdlEnum":
            return "IdlEnum"
        return f"{cls.__name__}(IdlEnum, idl_typename='{cls.__idl_typename__}')"



"""
 * Copyright(c) 2021 to 2022 ZettaScale Technology and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
"""

from math import log2
from enum import Enum
from dataclasses import dataclass

from .types import _type_code_align_size_default_mapping
from ._support import Buffer, CdrKeyVmOp, CdrKeyVMOpType, KeyScanner, SerializeKind, DeserializeKind
from ._support import XCDR1Constants as XCDR1
from . import types as types

# Note: the clayer always runs "normalize" on the serialized representation when a sample is constructed,
# so the Python-based deserializers only see well-formed serialized representations.

class KeyEnabled(Enum):
    Never = 0
    InKeylist = 1
    InKeylistOrKeyless = 2

class Machine:
    """Given a type, serialize and deserialize"""
    def __init__(self, type):
        self.alignment = 1

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        pass

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        pass

    def key_scan(self) -> KeyScanner:
        pass

    def cdr_key_machine_op(self, skip):
        pass

    def default_initialize(self):
        pass


class NoneMachine(Machine):
    def __init__(self):
        self.alignment = 1

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        pass

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        pass

    def key_scan(self) -> KeyScanner:
        return KeyScanner()

    def cdr_key_machine_op(self, skip):
        return []

    def default_initialize(self):
        return None


class PrimitiveMachine(Machine):
    def __init__(self, type):
        self.type = type
        self.code, self.alignment, self.size, self.default = _type_code_align_size_default_mapping[type]

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(self.alignment)
        buffer.write(self.code, self.size, value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(self.alignment)
        return buffer.read(self.code, self.size)

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(self.alignment, self.size)

    def cdr_key_machine_op(self, skip):
        stream = [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, self.size, align=self.alignment)]
        if not skip and not self.size == 1:
            # Note: in the byteswap op the align field indicates the byteswap size
            #       so the align=size is right, it is weird, agreed.
            stream += [CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=self.size)]
        return stream

    def default_initialize(self):
        return self.default


class CharMachine(Machine):
    def __init__(self):
        self.alignment = 1

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.write('b', 1, ord(value))

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        return chr(buffer.read('b', 1))

    def cdr_key_machine_op(self, skip):
        return [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, 1, align=1)]

    def default_initialize(self):
        return '\0'

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(1, 1)


class StringMachine(Machine):
    def __init__(self, bound=None):
        self.alignment = 4
        self.bound = bound

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.bound and len(value) > self.bound:
            raise Exception("String longer than bound.")
        buffer.align(4)
        bytes = value.encode('utf-8')
        buffer.write('I', 4, len(bytes) + 1)
        buffer.write_bytes(bytes)
        buffer.write('b', 1, 0)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)
        numbytes = buffer.read('I', 4)
        bytes = buffer.read_bytes(numbytes - 1)
        buffer.read('b', 1)
        return bytes.decode('utf-8')

    def key_scan(self) -> KeyScanner:
        if self.bound:
            return KeyScanner.with_bound(4, self.bound + 4)
        else:
            return KeyScanner.infinity()

    def cdr_key_machine_op(self, skip):
        return [CdrKeyVmOp(CdrKeyVMOpType.Stream4ByteSize, skip, 1, align=1)]

    def default_initialize(self):
        return ""


class BytesMachine(Machine):
    def __init__(self, bound=None):
        self.alignment = 4
        self.bound = bound

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.bound and len(value) > self.bound:
            raise Exception("Bytes longer than bound.")
        buffer.align(4)
        buffer.write('I', 4, len(value))
        buffer.write_bytes(value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)
        numbytes = buffer.read('I', 4)
        return buffer.read_bytes(numbytes)

    def key_scan(self) -> KeyScanner:
        if self.bound:
            return KeyScanner.with_bound(4, self.bound + 4)
        else:
            return KeyScanner.infinity()

    def cdr_key_machine_op(self, skip):
        return [CdrKeyVmOp(CdrKeyVMOpType.Stream4ByteSize, skip, 1, align=1)]

    def default_initialize(self):
        return bytes(0)


class ByteArrayMachine(Machine):
    def __init__(self, size):
        self.alignment = 1
        self.size = size

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if len(value) != self.size:
            raise Exception("Incorrectly sized array.")

        buffer.write_bytes(value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        return buffer.read_bytes(self.size)

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(1, self.size)

    def cdr_key_machine_op(self, skip):
        return [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, self.size, align=1)]

    def default_initialize(self):
        return bytearray(self.size)


class ArrayMachine(Machine):
    def __init__(self, submachine, size, add_size_header=False):
        self.size = size
        self.submachine = submachine
        self.alignment = submachine.alignment
        self.add_size_header = add_size_header

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        assert len(value) == self.size

        if self.add_size_header:
            buffer.align(4)
            buffer.write('I', 4, 0)
            hpos = buffer.tell()

        for v in value:
            self.submachine.serialize(buffer, v, serialize_kind, key_enabled)

        if self.add_size_header:
            mpos = buffer.tell()
            buffer.seek(hpos - 4)
            buffer.write('I', 4, mpos - hpos)
            buffer.seek(mpos)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.add_size_header:
            buffer.align(4)
            size = buffer.read('I', 4)
            mpos = buffer.tell()

        v = [self.submachine.deserialize(buffer, deserialize_kind, key_enabled) for i in range(self.size)]

        if self.add_size_header:
            assert buffer.tell() == mpos + size

        return v

    def key_scan(self) -> KeyScanner:
        scan = KeyScanner()
        scan.increase_by_multiplied_subresult(self.submachine.key_scan(), self.size)
        return scan

    def cdr_key_machine_op(self, skip):
        if skip and self.add_size_header:
            return [CdrKeyVmOp(CdrKeyVMOpType.Stream4ByteSize, True, size=1, align=4)]

        subops = self.submachine.cdr_key_machine_op(skip)
        return ([CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, size=4, align=4)] if self.add_size_header else []) + \
            ([CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=4)] if not skip and self.add_size_header else []) + \
            [CdrKeyVmOp(CdrKeyVMOpType.RepeatStatic, skip, self.size, value=len(subops) + 2)] + \
            subops + [CdrKeyVmOp(CdrKeyVMOpType.EndRepeat, skip, len(subops))]

    def default_initialize(self):
        return [self.submachine.default_initialize() for i in range(self.size)]


class SequenceMachine(Machine):
    def __init__(self, submachine, maxlen=None, add_size_header=False):
        self.submachine = submachine
        self.alignment = 4
        self.maxlen = maxlen
        self.add_size_header = add_size_header

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.maxlen is not None:
            assert len(value) <= self.maxlen

        buffer.align(4)

        if self.add_size_header:
            buffer.write('I', 4, 0)
            hpos = buffer.tell()

        buffer.write('I', 4, len(value))

        for v in value:
            self.submachine.serialize(buffer, v, serialize_kind, key_enabled)

        if self.add_size_header:
            mpos = buffer.tell()
            buffer.seek(hpos - 4)
            buffer.write('I', 4, mpos - hpos)
            buffer.seek(mpos)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)

        if self.add_size_header:
            size = buffer.read('I', 4)
            mpos = buffer.tell()

        num = buffer.read('I', 4)
        v = [self.submachine.deserialize(buffer, deserialize_kind, key_enabled) for i in range(num)]

        if self.add_size_header:
            buffer.seek(mpos + size)

        return v

    def key_scan(self) -> KeyScanner:
        if not self.maxlen:
            return KeyScanner.infinity()

        if self.add_size_header:
            scan = KeyScanner.with_bound(4, 8)
        else:
            scan = KeyScanner.with_bound(4, 4)

        scan.increase_by_multiplied_subresult(self.submachine.key_scan(), self.maxlen)

        if scan.size > 1_000_000:
            return KeyScanner.infinity()
        return scan

    def cdr_key_machine_op(self, skip):
        if skip and self.add_size_header:
            return [CdrKeyVmOp(CdrKeyVMOpType.Stream4ByteSize, True, size=1, align=4)]

        subops = self.submachine.cdr_key_machine_op(skip)
        return ([CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, size=4, align=4)] if self.add_size_header else []) + \
            ([CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=4)] if not skip and self.add_size_header else []) + \
            [CdrKeyVmOp(CdrKeyVMOpType.Repeat4ByteSize, skip, value=len(subops) + 2)] + \
            subops + [CdrKeyVmOp(CdrKeyVMOpType.EndRepeat, skip, len(subops))]

    def default_initialize(self):
        return []


class UnionMachine(Machine):
    def __init__(self, type, discriminator_machine, labels_submachines, default_case=None):
        self.type = type
        self.labels_submachines = labels_submachines
        self.alignment = max(s.alignment for s in labels_submachines.values())
        self.alignment = max(self.alignment, discriminator_machine.alignment)
        self.discriminator = discriminator_machine
        self.default = default_case
        self.discriminator_is_key = type.__idl_annotations__.get("discriminator_is_key", False)

    def serialize(self, buffer, union, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        discr, value = union.get()

        if (serialize_kind == SerializeKind.KeyDefinitionOrder or serialize_kind == SerializeKind.KeyNormalized) and self.discriminator_is_key:
            try:
                if discr is None:
                    self.discriminator.serialize(buffer, union.__idl_default_discriminator__, serialize_kind, key_enabled)
                else:
                    self.discriminator.serialize(buffer, discr, serialize_kind, key_enabled)
                return
            except Exception as e:
                raise Exception(f"Failed to encode union, {self.type}, value is {value}, discriminator is {discr}") from e

        try:
            if discr is None:
                self.discriminator.serialize(buffer, union.__idl_default_discriminator__, serialize_kind, key_enabled)
                if self.default:
                    self.default.serialize(buffer, value, serialize_kind, KeyEnabled.Never)
            elif discr not in self.labels_submachines:
                self.discriminator.serialize(buffer, discr, serialize_kind, key_enabled)
                if self.default:
                    self.default.serialize(buffer, value, serialize_kind, KeyEnabled.Never)
            else:
                self.discriminator.serialize(buffer, discr, serialize_kind, key_enabled)
                self.labels_submachines[discr].serialize(buffer, value, serialize_kind, KeyEnabled.Never)
        except Exception as e:
            raise Exception(f"Failed to encode union, {self.type}, value is {value}, discriminator is {discr}") from e

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        label = self.discriminator.deserialize(buffer, deserialize_kind, key_enabled)

        if label not in self.labels_submachines:
            if self.default:
                contents = self.default.deserialize(buffer, deserialize_kind, KeyEnabled.Never)
            else:
                contents = None
        else:
            contents = self.labels_submachines[label].deserialize(buffer, deserialize_kind, KeyEnabled.Never)

        return self.type(discriminator=label, value=contents)

    def key_scan(self) -> KeyScanner:
        dscan = self.discriminator.key_scan()
        if self.discriminator_is_key:
            return dscan

        scan = KeyScanner()
        for _, machine in self.labels_submachines.items():
            scan = scan.max(machine.key_scan())

        dscan.increase_by_multiplied_subresult(scan, 1)
        return dscan

    def cdr_key_machine_op(self, skip):  # TODO: check again
        headers = []
        opsets = []
        union_type = {
            1: CdrKeyVMOpType.Union1Byte,
            2: CdrKeyVMOpType.Union2Byte,
            4: CdrKeyVMOpType.Union4Byte,
            8: CdrKeyVMOpType.Union8Byte
        }[self.discriminator.alignment]

        buffer = Buffer(_bytes=self.discriminator.alignment)

        value_skip = skip or self.type.__idl_discriminator_is_key__

        for label, submachine in self.labels_submachines.items():
            buffer.seek(0)
            self.discriminator.serialize(buffer, label)
            buffer.seek(0)
            value = buffer.read({1: 'B', 2: 'H', 4: 'I', 8: 'Q'}[self.discriminator.alignment], self.discriminator.alignment)
            headers.append(CdrKeyVmOp(union_type, skip, value=value))
            opsets.append(submachine.cdr_key_machine_op(value_skip))

        lens = [len(o) + 2 for o in opsets]

        if self.default is not None:
            opsets.append(self.discriminator.cdr_key_machine_op(skip) + self.default.cdr_key_machine_op(value_skip))
            lens.append(len(opsets[-1]))
        else:
            lens[-1] -= 1

        jumps = [sum(lens[i:]) for i in range(len(lens))]

        for i in range(len(headers)):
            if i != len(opsets) - 1:
                opsets[i].append(CdrKeyVmOp(CdrKeyVMOpType.Jump, skip, size=jumps[i + 1] + 1))
            headers[i].size = lens[i]
            opsets[i] = [headers[i]] + opsets[i]

        return sum(opsets, [])

    def default_initialize(self):
        return self.type(
            discriminator=self.type.__idl_default_discriminator__,
            value=None if self.default is None else self.default.default_initialize()
        )


class MappingMachine(Machine):
    def __init__(self, key_machine, value_machine):
        self.key_machine = key_machine
        self.value_machine = value_machine
        self.alignment = 4

    def serialize(self, buffer, values, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)
        buffer.write('I', 4, len(values))

        for key, value in values.items():
            self.key_machine.serialize(buffer, key, serialize_kind, key_enabled)
            self.value_machine.serialize(buffer, value, serialize_kind, key_enabled)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        ret = {}
        buffer.align(4)
        num = buffer.read('I', 4)

        for _i in range(num):
            key = self.key_machine.deserialize(buffer, deserialize_kind, key_enabled)
            value = self.value_machine.deserialize(buffer, deserialize_kind, key_enabled)
            ret[key] = value

        return ret

    def key_scan(self) -> KeyScanner:
        return KeyScanner.infinity()

    def cdr_key_machine_op(self, skip):
        raise NotImplementedError()

    def default_initialize(self):
        return {}


class StructMachine(Machine):
    def __init__(self, object, members_machines, keylist):
        self.type = object
        self.members_machines = members_machines
        self.keylist = keylist

    def key_enabled(self, member, key_enabled):
        if key_enabled == KeyEnabled.Never:
            return KeyEnabled.Never
        elif key_enabled == KeyEnabled.InKeylist:
            return KeyEnabled.InKeylistOrKeyless if self.keylist and member in self.keylist else KeyEnabled.Never
        else:
            assert key_enabled == KeyEnabled.InKeylistOrKeyless
            return KeyEnabled.InKeylistOrKeyless if not self.keylist or member in self.keylist else KeyEnabled.Never

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        #  We use the fact here that dicts retain their insertion order
        #  This is guaranteed from python 3.7
        for member, machine in self.members_machines.items():
            m_key_enabled = self.key_enabled(member, key_enabled)
            if serialize_kind != SerializeKind.DataSample and m_key_enabled == KeyEnabled.Never:
                continue

            try:
                machine.serialize(buffer, getattr(value, member), serialize_kind, m_key_enabled)
            except Exception as e:
                raise Exception(f"Failed to encode member {member}, value is {getattr(value, member)}") from e

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        valuedict = {}
        for member, machine in self.members_machines.items():
            m_key_enabled = self.key_enabled(member, key_enabled)
            if deserialize_kind != DeserializeKind.DataSample and m_key_enabled == KeyEnabled.Never:
                valuedict[member] = machine.default_initialize()
            else:
                valuedict[member] = machine.deserialize(buffer, deserialize_kind, m_key_enabled)
        return self.type(**valuedict)

    def key_scan(self) -> KeyScanner:
        scan = KeyScanner()

        for member, machine in self.members_machines.items():
            if self.keylist and member not in self.keylist:
                continue
            scan.increase_by_multiplied_subresult(machine.key_scan(), 1)

        return scan

    def cdr_key_machine_op(self, skip):
        return sum(
            (
                m.cdr_key_machine_op(skip or (self.keylist and name not in self.keylist))
                for name, m in self.members_machines.items()
            ),
            []
        )

    def default_initialize(self):
        valuedict = {}
        for member, machine in self.members_machines.items():
            valuedict[member] = machine.default_initialize()
        return self.type(**valuedict)


class InstanceMachine(Machine):
    def __init__(self, object, use_version_2):
        self.type = object
        self.alignment = 1
        self.use_version_2 = use_version_2

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.type.__idl__.v1_machine is None:
            self.type.__idl__.populate()

        if self.use_version_2:
            return self.type.__idl__.v2_machine.serialize(buffer, value, serialize_kind, key_enabled)
        else:
            return self.type.__idl__.v1_machine.serialize(buffer, value, serialize_kind, key_enabled)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.type.__idl__.v1_machine is None:
            self.type.__idl__.populate()

        if self.use_version_2:
            return self.type.__idl__.v2_machine.deserialize(buffer, deserialize_kind, key_enabled)
        else:
            return self.type.__idl__.v1_machine.deserialize(buffer, deserialize_kind, key_enabled)

    def key_scan(self):
        return self.type.__idl__.key_scan(use_version_2=self.use_version_2)

    def cdr_key_machine_op(self, skip):
        return self.type.__idl__.cdr_key_machine(skip, use_version_2=self.use_version_2)

    def default_initialize(self):
        if self.type.__idl__.v1_machine is None:
            self.type.__idl__.populate()

        if self.use_version_2:
            return self.type.__idl__.v2_machine.default_initialize()
        else:
            return self.type.__idl__.v1_machine.default_initialize()


class EnumMachine(Machine):
    def __init__(self, enum):
        self.enum: Enum = enum
        self.alignment = 4
        self.size = 4

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)
        if type(value) == int:
            buffer.write("I", 4, value)
            return
        buffer.write("I", 4, value.value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)
        v = buffer.read("I", 4)
        try:
            return self.enum(v)
        except ValueError:
            return v

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(4, 4)

    def cdr_key_machine_op(self, skip):
        stream = [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, 4, align=4)]
        if not skip:
            stream += [CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=4)]
        return stream

    def default_initialize(self):
        return self.enum.__idl_enum_default_value__


class BitBoundEnumMachine(Machine):
    def __init__(self, enum, bit_bound):
        self.bit_bound = bit_bound
        self.encoder = [types.uint8, types.uint16, types.uint32, types.uint64][int(log2(bit_bound)) - 3]
        self.enum: Enum = enum
        self.code, self.alignment, self.size, _ = _type_code_align_size_default_mapping[self.encoder]

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if type(value) == int:
            buffer.align(self.alignment)
            buffer.write(self.code, self.size, value)
            return
        buffer.align(self.alignment)
        buffer.write(self.code, self.size, value.value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(self.alignment)
        v = buffer.read(self.code, self.size)
        try:
            return self.enum(v)
        except ValueError:
            return v

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(self.alignment, self.size)

    def cdr_key_machine_op(self, skip):
        stream = [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, self.size, align=self.alignment)]
        if not skip:
            stream += [CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=self.alignment)]
        return stream

    def default_initialize(self):
        return self.enum.__idl_enum_default_value__


class OptionalMachine(Machine):
    def __init__(self, submachine, memberid_muflag, use_version_2):
        self.submachine = submachine
        self.memberid_muflag = memberid_muflag
        self.use_version_2 = use_version_2

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        assert not (serialize_kind == SerializeKind.KeyDefinitionOrder or serialize_kind == SerializeKind.KeyNormalized)
        if self.use_version_2:
            if value is None:
                buffer.write('?', 1, False)
            else:
                buffer.write('?', 1, True)
                self.submachine.serialize(buffer, value, serialize_kind, KeyEnabled.Never)
        else:
            # TODO: compact variant if it fits
            buffer.align(4)
            buffer.write('H', 2, XCDR1.PL_SHORT_PID_EXTENDED | XCDR1.PL_SHORT_FLAG_MU)
            buffer.write('H', 2, XCDR1.PL_SHORT_PID_EXT_LEN)
            buffer.write('I', 4, self.memberid_muflag)
            hpos = buffer.tell()
            buffer.write('I', 4, 0)
            dpos = buffer.tell()
            if value is not None:
                old_align_offset = buffer.set_align_offset(dpos)
                self.submachine.serialize(buffer, value, serialize_kind, KeyEnabled.Never)
                buffer.set_align_offset(old_align_offset)
                fpos = buffer.tell()
                buffer.seek(hpos)
                buffer.write('I', 4, fpos - dpos)
                buffer.seek(fpos)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.use_version_2:
            if buffer.read('?', 1):
                return self.submachine.deserialize(buffer, deserialize_kind, KeyEnabled.Never)
            return None
        else:
            buffer.align(4)
            header = buffer.read('H', 2)
            membersize = buffer.read('H', 2)
            if (header & XCDR1.PL_SHORT_PID_MASK) == XCDR1.PL_SHORT_PID_EXTENDED:
                # extended form; memberlen should be XCDR1.PL_SHORT_PID_EXT_LEN
                header = buffer.read('I', 4)
                membersize = buffer.read('I', 4)
            # header & XCDR1.PL_SHORT_PID_MASK should be self.memberid, here we assume it is
            # (C code checks it in "normalize", which always runs before we get here)
            if membersize == 0:
                return None
            else:
                old_align_offset = buffer.set_align_offset(buffer.tell())
                val = self.submachine.deserialize(buffer, deserialize_kind, KeyEnabled.Never)
                buffer.set_align_offset(old_align_offset)
                return val

    def key_scan(self) -> KeyScanner:
        scan = KeyScanner.simple(1, 1)
        scan.increase_by_multiplied_subresult(self.submachine.key_scan(), 1)
        return scan

    def cdr_key_machine_op(self, skip):
        subops = self.submachine.cdr_key_machine_op(skip)
        return [CdrKeyVmOp(CdrKeyVMOpType.Optional, skip, len(subops) + 1, align=1)] + subops

    def default_initialize(self):
        return None


class PlainCdrV2ArrayOfPrimitiveMachine(Machine):
    def __init__(self, type, length):
        code, self.alignment, size, default = types._type_code_align_size_default_mapping[type]
        self.length = length
        self.size = size * length
        self.code = str(length) + code
        self.default = [default] * length
        self.subtype = type

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        assert len(value) == self.length
        buffer.align(self.alignment)
        buffer.write_multi(self.code, self.size, *value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(self.alignment)
        return list(buffer.read_multi(self.code, self.size))

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(self.alignment, self.size)

    def cdr_key_machine_op(self, skip):
        stream = [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, self.size, align=self.alignment)]
        if not skip and not self.alignment == 1:
            stream += [CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=self.alignment)]
        return stream

    def default_initialize(self):
        return self.default.copy()


class PlainCdrV2SequenceOfPrimitiveMachine(Machine):
    def __init__(self, type, max_length=None):
        self.code, self.alignment, self.size, _ = types._type_code_align_size_default_mapping[type]
        self.max_length = max_length

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        assert self.max_length is None or len(value) <= self.max_length
        buffer.align(4)
        buffer.write('I', 4, len(value))
        if value:
            buffer.align(self.alignment)
            buffer.write_multi(f"{len(value)}{self.code}", self.size * len(value), *value)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(4)
        length = buffer.read('I', 4)
        if length:
            buffer.align(self.alignment)
            return list(buffer.read_multi(f"{length}{self.code}", self.size * length))
        else:
            return []

    def key_scan(self) -> KeyScanner:
        if not self.max_length:
            return KeyScanner.infinity()

        scan = KeyScanner.with_bound(4, 4)
        scan.increase_by_multiplied_subresult(
            KeyScanner.simple(self.size, self.size),
            self.max_length
        )

        if scan.size > 1_000_000:
            return KeyScanner.infinity()
        return scan

    def cdr_key_machine_op(self, skip):
        stream = [CdrKeyVmOp(
            CdrKeyVMOpType.Stream4ByteSize,
            skip, self.size, align=self.alignment
        )]
        if not skip and self.size != 1:
            stream += [CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=self.size)]
        return stream

    def default_initialize(self):
        return []


class DelimitedCdrAppendableStructMachine(Machine):
    def __init__(self, type, member_machines, keylist):
        self.alignment = 4
        self.type = type
        self.member_machines = member_machines
        self.keylist = keylist

    def key_enabled(self, member, key_enabled):
        if key_enabled == KeyEnabled.Never:
            return KeyEnabled.Never
        elif key_enabled == KeyEnabled.InKeylist:
            return KeyEnabled.InKeylistOrKeyless if self.keylist and member in self.keylist else KeyEnabled.Never
        else:
            assert key_enabled == KeyEnabled.InKeylistOrKeyless
            return KeyEnabled.InKeylistOrKeyless if not self.keylist or member in self.keylist else KeyEnabled.Never

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        # write dummy header
        buffer.align(4)
        hpos = buffer.tell()
        buffer.write('I', 4, 0)

        # write member data
        dpos = buffer.tell()
        for member, machine in self.member_machines.items():
            m_key_enabled = self.key_enabled(member, key_enabled)
            if serialize_kind != SerializeKind.DataSample and m_key_enabled == KeyEnabled.Never:
                continue

            try:
                machine.serialize(buffer, getattr(value, member), serialize_kind, m_key_enabled)
            except Exception as e:
                raise Exception(f"Failed to encode member {member}, value is {getattr(value, member)}") from e

        fpos = buffer.tell()

        # Write size header word back
        buffer.seek(hpos)
        buffer.write('I', 4, fpos - dpos)
        buffer.seek(fpos)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        # read header
        buffer.align(4)
        size = buffer.read('I', 4)
        hpos = buffer.tell()

        data = {}
        for member, machine in self.member_machines.items():
            m_key_enabled = self.key_enabled(member, key_enabled)
            if buffer.tell() - hpos >= size:
                data[member] = machine.default_initialize()
            elif deserialize_kind != DeserializeKind.DataSample and m_key_enabled == KeyEnabled.Never:
                data[member] = machine.default_initialize()
            else:
                data[member] = machine.deserialize(buffer, deserialize_kind, m_key_enabled)

            if buffer.tell() - hpos > size:
                raise Exception("Struct was not contained inside header indicated size, stream corrupt.")

        buffer.seek(hpos + size)
        return self.type(**data)

    def key_scan(self) -> KeyScanner:
        scan = KeyScanner()

        for member, machine in self.member_machines.items():
            if self.keylist and member not in self.keylist:
                continue
            scan.increase_by_multiplied_subresult(machine.key_scan(), 1)

        return scan

    def cdr_key_machine_op(self, skip):
        if skip:
            # This sub-struct contains nothing interesting, we can skip it
            return [CdrKeyVmOp(CdrKeyVMOpType.StructHeader, skip=skip, size=0, value=0, align=4)]

        members = sum(
            (
                m.cdr_key_machine_op(self.keylist and name not in self.keylist)
                for name, m in self.member_machines.items()
            ),
            []
        )
        while members and members[-1].skip:
            members.pop()

        return [CdrKeyVmOp(CdrKeyVMOpType.AppendableHeader, False)] + \
            members + \
            [CdrKeyVmOp(CdrKeyVMOpType.AppendableJumpToEnd, False)]

    def default_initialize(self):
        valuedict = {}
        for member, machine in self.member_machines.items():
            valuedict[member] = machine.default_initialize()
        return self.type(**valuedict)


class DelimitedCdrAppendableUnionMachine(Machine):
    def __init__(self, type, discriminator_machine, labels_submachines, default_case=None):
        self.type = type
        self.labels_submachines = labels_submachines
        self.alignment = 4
        self.discriminator = discriminator_machine
        self.default = default_case
        self.discriminator_is_key = type.__idl_annotations__.get("discriminator_is_key", False)

    def serialize(self, buffer, union, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        # write dummy header
        buffer.align(4)
        hpos = buffer.tell()
        buffer.write('I', 4, 0)

        dpos = buffer.tell()
        discr, value = union.get()

        if (serialize_kind == SerializeKind.KeyDefinitionOrder or serialize_kind == SerializeKind.KeyNormalized) and self.discriminator_is_key:
            try:
                if discr is None:
                    self.discriminator.serialize(buffer, union.__idl_default_discriminator__, serialize_kind)
                else:
                    self.discriminator.serialize(buffer, discr, serialize_kind)
                return
            except Exception as e:
                raise Exception(f"Failed to encode union, {self.type}, value is {value}") from e

        try:
            if discr is None:
                self.discriminator.serialize(buffer, union.__idl_default_discriminator__)
                if self.default:
                    self.default.serialize(buffer, value, serialize_kind)
            elif discr not in self.labels_submachines:
                self.discriminator.serialize(buffer, discr, serialize_kind)
                if self.default:
                    self.default.serialize(buffer, value, serialize_kind)
            else:
                self.discriminator.serialize(buffer, discr, serialize_kind)
                self.labels_submachines[discr].serialize(buffer, value, serialize_kind)
        except Exception as e:
            raise Exception(f"Failed to encode union, {self.type}, value is {value}") from e

        fpos = buffer.tell()

        # Write size header word back
        buffer.seek(hpos)
        buffer.write('I', 4, fpos - dpos)
        buffer.seek(fpos)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        # read header
        buffer.align(4)
        size = buffer.read('I', 4)
        hpos = buffer.tell()

        label = self.discriminator.deserialize(buffer, deserialize_kind)
        lpos = buffer.tell()

        if label not in self.labels_submachines:
            if lpos == hpos + size:
                return self.default_initialize()

            if self.default:
                contents = self.default.deserialize(buffer, deserialize_kind)
            else:
                contents = None
        else:
            if lpos == hpos + size:
                contents = self.labels_submachines[label].default_initialize()
            else:
                contents = self.labels_submachines[label].deserialize(buffer, deserialize_kind)

        buffer.seek(hpos + size)
        return self.type(discriminator=label, value=contents)

    def key_scan(self) -> KeyScanner:
        dscan = self.discriminator.key_scan()
        if self.discriminator_is_key:
            return dscan

        scan = KeyScanner()
        for _, machine in self.labels_submachines.items():
            scan = scan.max(machine.key_scan())

        dscan.increase_by_multiplied_subresult(scan, 1)
        return dscan

    def cdr_key_machine_op(self, skip):  # TODO: check again
        if skip:
            # This union contains nothing interesting, we can skip it
            return [CdrKeyVmOp(CdrKeyVMOpType.StructHeader, skip=skip, size=0, value=0, align=4)]

        headers = []
        opsets = []
        union_type = {
            1: CdrKeyVMOpType.Union1Byte,
            2: CdrKeyVMOpType.Union2Byte,
            4: CdrKeyVMOpType.Union4Byte,
            8: CdrKeyVMOpType.Union8Byte
        }[self.discriminator.alignment]

        buffer = Buffer(_bytes=self.discriminator.alignment)

        value_skip = skip or self.type.__idl_discriminator_is_key__

        for label, submachine in self.labels_submachines.items():
            buffer.seek(0)
            self.discriminator.serialize(buffer, label)
            buffer.seek(0)
            value = buffer.read({1: 'B', 2: 'H', 4: 'I', 8: 'Q'}[self.discriminator.alignment], self.discriminator.alignment)
            headers.append(CdrKeyVmOp(union_type, skip, value=value))
            opsets.append(submachine.cdr_key_machine_op(value_skip))

        lens = [len(o) + 2 for o in opsets]

        if self.default is not None:
            opsets.append(self.discriminator.cdr_key_machine_op(skip) + self.default.cdr_key_machine_op(value_skip))
            lens.append(len(opsets[-1]))
        else:
            lens[-1] -= 1

        jumps = [sum(lens[i:]) for i in range(len(lens))]

        for i in range(len(headers)):
            if i != len(opsets) - 1:
                opsets[i].append(CdrKeyVmOp(CdrKeyVMOpType.Jump, skip, size=jumps[i + 1] + 1))
            headers[i].size = lens[i]
            opsets[i] = [headers[i]] + opsets[i]

        return [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip=True, size=4, align=4)] + sum(opsets, [])

    def default_initialize(self):
        return self.type(
            discriminator=self.type.__idl_default_discriminator__,
            value=None if self.default is None else self.default.default_initialize()
        )


# Todo: Mutable unions

class LenType(Enum):
    OneByte = 0
    TwoByte = 1
    FourByte = 2
    EightByte = 3
    NextIntLen = 4
    NextIntDualUseLen = 5
    NextIntDualUse4Len = 6
    NextIntDualUse8Len = 7


@dataclass
class MutableMember:
    name: str
    key: bool
    optional: bool
    lentype: LenType
    must_understand: bool
    memberid: int
    machine: Machine
    header: int = 0

    def __post_init__(self):
        # "must understand" also gets set for key fields, but whether a field is a key
        # field or not is dependent on the annotation and the parent, and therefore more
        # easily done during serialization
        self.header = ((1 if self.must_understand else 0) << 31) + (self.lentype.value << 28) + (self.memberid & 0x0fffffff)


class MustUnderstandFailure(Exception):
    pass

class KeyFieldNotProvidedFailure(Exception):
    pass

class PLCdrMutableStructMachine(Machine):
    def __init__(self, type, mutablemembers, use_version_2):
        self.alignment = 4
        self.type = type
        self.mutablemembers = mutablemembers
        self.use_version_2 = use_version_2
        self.keylist = [m for m in self.mutablemembers if m.key]
        self.mutmem_by_id = {
            m.memberid: m for m in mutablemembers
        }
        self.init_map = {
            m.name: None for m in mutablemembers
        }

    def key_enabled(self, member, key_enabled):
        if key_enabled == KeyEnabled.Never:
            return KeyEnabled.Never
        elif key_enabled == KeyEnabled.InKeylist:
            return KeyEnabled.InKeylistOrKeyless if member.key else KeyEnabled.Never
        else:
            assert key_enabled == KeyEnabled.InKeylistOrKeyless
            return KeyEnabled.InKeylistOrKeyless if not self.keylist or member.key else KeyEnabled.Never

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.use_version_2:
            buffer.align(4)
            hpos = buffer.tell()
            buffer.write('I', 4, 0)

        # write member data
        dpos = buffer.tell()
        members = sorted(self.mutablemembers, key=lambda x: x.memberid) if serialize_kind == SerializeKind.KeyNormalized else self.mutablemembers
        for mutablemember in members:
            m_key_enabled = self.key_enabled(mutablemember, key_enabled)
            if serialize_kind != SerializeKind.DataSample and m_key_enabled == KeyEnabled.Never:
                continue

            member_value = getattr(value, mutablemember.name)
            if mutablemember.optional and member_value is None:
                continue

            buffer.align(4)
            if self.use_version_2:
                # It appears RTI also can't handle "must understand" on key fields in
                # XCDR2, even though XTypes 1.3 requires it.  This should add:
                #   | ((1 if m_key_enabled != KeyEnabled.Never else 0) << 31)
                buffer.write('I', 4, mutablemember.header)
            else:
                # TODO: use compact variant when member id and the max serialized size of the type are small enough
                buffer.write('H', 2, XCDR1.PL_SHORT_PID_EXTENDED | XCDR1.PL_SHORT_FLAG_MU)
                buffer.write('H', 2, XCDR1.PL_SHORT_PID_EXT_LEN)
                # It appears RTI can't handle "must understand" on key fields, even though
                # XTypes 1.3 requires it.  This should read:
                #   mu_flag = (1 if mutablemember.must_understand or
                #              m_key_enabled != KeyEnabled.Never else 0) << 30
                mu_flag = (1 if mutablemember.must_understand else 0) << 30
                buffer.write('I', 4, mu_flag | mutablemember.memberid)
                assert mutablemember.lentype == LenType.NextIntLen

            mpos = buffer.tell()
            if mutablemember.lentype == LenType.NextIntLen:
                buffer.write('I', 4, 0)

            old_align_offset = None
            if not self.use_version_2:
                old_align_offset = buffer.set_align_offset(buffer.tell())

            mutablemember.machine.serialize(buffer, member_value, serialize_kind, m_key_enabled)

            if not self.use_version_2:
                buffer.set_align_offset(old_align_offset)
                # RTI implements/prefers/requires XTypes 1.1-style XCDR1 with the lengths padded to a
                # multiple of 4. XTypes 1.3 requires the lengths to not be padded. There's no good way
                # out ...
                buffer.align(4)

            if mutablemember.lentype == LenType.NextIntLen:
                ampos = buffer.tell()
                buffer.seek(mpos)
                buffer.write('I', 4, ampos - mpos - 4)
                buffer.seek(ampos)

        if self.use_version_2:
            # Write size header word back
            fpos = buffer.tell()
            buffer.seek(hpos)
            buffer.write('I', 4, fpos - dpos)
            buffer.seek(fpos)
        else:
            # Write sentinel
            buffer.align(4)
            buffer.write('H', 2, XCDR1.PL_SHORT_PID_LIST_END | XCDR1.PL_SHORT_FLAG_MU)
            buffer.write('H', 2, 0)

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        if self.use_version_2:
            # read header
            buffer.align(4)
            struct_size = buffer.read('I', 4)
        else:
            # a lie, but it helps with the loop needed for XCDR2
            struct_size = 0xffffffff
        hpos = buffer.tell()

        data = self.init_map.copy()
        while buffer.tell() - hpos < struct_size:
            buffer.align(4)
            if self.use_version_2:
                header = buffer.read('I', 4)
                must_understand = ((header >> 31) & 1) > 0
                memberid = header & 0x0fffffff
                lc = (header >> 28) & 0x7
                if lc < 4:
                    membersize = 2 ** lc
                elif lc == 4: # "nextint": read length
                    membersize = buffer.read('I', 4)
                else: # "also nextint": peek length
                    lpos = buffer.tell()
                    membersize = buffer.read('I', 4)
                    buffer.seek(lpos)
                    if lc == 6:
                        membersize *= 4
                    elif lc == 7:
                        membersize *= 8
                    # extra 4 bytes for length
                    membersize += 4
            else:
                header = buffer.read('H', 2)
                membersize = buffer.read('H', 2)
                if (header & XCDR1.PL_SHORT_PID_MASK) == XCDR1.PL_SHORT_PID_LIST_END:
                    break # sentinel ends XCDRv1 list
                if (header & XCDR1.PL_SHORT_PID_MASK) == XCDR1.PL_SHORT_PID_EXTENDED:
                    # long form; memberlen should be XCDR1.PL_SHORT_PID_EXT_LEN
                    header = buffer.read('I', 4)
                    membersize = buffer.read('I', 4)
                    must_understand = ((header >> 30) & 1) > 0
                    memberid = header & 0x0fffffff
                else:
                    must_understand = (header & XCDR1.PL_SHORT_FLAG_MU) > 0
                    memberid = (header & XCDR1.PL_SHORT_PID_MASK)
                    if (header & XCDR1.PL_SHORT_FLAG_IMPL_EXT) > 0:
                        memberid ^= XCDR1.PL_SHORT_FLAG_IMPL_EXT | XCDR1.PL_LONG_FLAG_IMPL_EXT

            mutmem = self.mutmem_by_id.get(memberid)

            mpos = buffer.tell()
            if mutmem:
                m_key_enabled = self.key_enabled(mutmem, key_enabled)
                if deserialize_kind == DeserializeKind.DataSample or m_key_enabled != KeyEnabled.Never:
                    if not self.use_version_2:
                        old_align_offset = buffer.set_align_offset(buffer.tell())
                    data[mutmem.name] = mutmem.machine.deserialize(buffer, deserialize_kind, m_key_enabled)
                    if not self.use_version_2:
                        buffer.set_align_offset(old_align_offset)
            else:
                if must_understand:
                    # Got a member that we don't know and marked as must understand: failure
                    raise MustUnderstandFailure()
            buffer.seek(mpos + membersize)

        for mutmem in self.mutablemembers:
            if data[mutmem.name] is None and mutmem.key:
                raise KeyFieldNotProvidedFailure()
            if data[mutmem.name] is None and not mutmem.optional:
                data[mutmem.name] = mutmem.machine.default_initialize()

        if self.use_version_2:
            buffer.seek(hpos + struct_size)
        return self.type(**data)

    def key_scan(self) -> KeyScanner:
        scan = KeyScanner()

        for m_id in sorted(self.mutmem_by_id.keys()):
            mutablemember = self.mutmem_by_id[m_id]
            if not mutablemember.key:
                continue
            scan.increase_by_multiplied_subresult(mutablemember.machine.key_scan(), 1)

        return scan

    def cdr_key_machine_op(self, skip):
        if skip:
            # This sub-struct contains nothing interesting, we can skip it
            return [CdrKeyVmOp(CdrKeyVMOpType.StructHeader, skip=skip, size=0, value=0, align=4)]

        ret = []
        for m_id in sorted(self.mutmem_by_id.keys()):
            mutmem = self.mutmem_by_id[m_id]
            if mutmem.key:
                ret += [
                    CdrKeyVmOp(CdrKeyVMOpType.MemberSelect, False, 0, m_id, 4)
                ] + mutmem.machine.cdr_key_machine_op(False) + [
                    CdrKeyVmOp(CdrKeyVMOpType.MemberSelectEnd, False, 0, m_id, 4)
                ]

        # position sample cursor at end of struct by 'skipping' from struct header
        ret += [CdrKeyVmOp(CdrKeyVMOpType.StructHeader, skip=True, size=0, value=0, align=4)]

        return ret

    def default_initialize(self):
        data = {}
        for mutmem in self.mutablemembers:
            if not mutmem.optional:
                data[mutmem.name] = mutmem.machine.default_initialize()
            else:
                data[mutmem.name] = None
        return self.type(**data)


class BitMaskMachine(Machine):
    def __init__(self, type, bit_bound=64):
        self.bit_bound = bit_bound
        self.type = type
        self.primitive_type = \
            types.uint8 if bit_bound < 9 else \
            types.uint16 if bit_bound < 17 else \
            types.uint32 if bit_bound < 33 else \
            types.uint64
        self.code, self.alignment, self.size, self.default = \
            types._type_code_align_size_default_mapping[self.primitive_type]

    def serialize(self, buffer, value, serialize_kind=SerializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(self.alignment)
        buffer.write(self.code, self.size, value.as_mask())

    def deserialize(self, buffer, deserialize_kind=DeserializeKind.DataSample, key_enabled=KeyEnabled.InKeylist):
        buffer.align(self.alignment)
        return self.type.from_mask(buffer.read(self.code, self.size))

    def key_scan(self) -> KeyScanner:
        return KeyScanner.simple(self.alignment, self.size)

    def cdr_key_machine_op(self, skip):
        stream = [CdrKeyVmOp(CdrKeyVMOpType.StreamStatic, skip, self.size, align=self.alignment)]
        if not skip and not self.size == 1:
            # Note: in the byteswap op the align field indicates the byteswap size
            #       so the align=size is right, it is weird, agreed.
            stream += [CdrKeyVmOp(CdrKeyVMOpType.ByteSwap, skip, align=self.size)]
        return stream

    def default_initialize(self):
        return self.default



import pytest

from dataclasses import dataclass

from cyclonedds.idl import IdlStruct, IdlBitmask, IdlUnion, IdlEnum
from cyclonedds.idl._support import Buffer, Endianness
import cyclonedds.idl._machinery as mc
import cyclonedds.idl.types as tp


class A(IdlEnum):
    V1 = 0xca
    V2 = 0xef


@dataclass
class B(IdlBitmask):
    V1: bool
    V2: bool


@dataclass
class C(IdlStruct):
    A: tp.uint8
    B: tp.uint16


class D(IdlUnion, discriminator=bool):
    A: tp.case[True, tp.uint16]
    B: tp.default[tp.int8]



def test_all_machine_serializers():
    b = Buffer()
    b.set_endianness(Endianness.Little)

    m = mc.CharMachine()
    m.serialize(b, "a")
    m.serialize(b, "b")
    assert b.asbytes() == b"ab"

    m = mc.PrimitiveMachine(tp.uint8)
    b.zero_out()
    b.seek(0)
    m.serialize(b, 0xab)
    assert b.asbytes() == b"\xab"

    m2 = mc.SequenceMachine(m, add_size_header=False)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, [0x0a, 0x0b])
    assert b.asbytes() == b"\x02\x00\x00\x00\x0a\x0b"

    m2 = mc.SequenceMachine(m, add_size_header=True)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, [0x0a, 0x0b])
    assert b.asbytes() == b"\x06\x00\x00\x00\x02\x00\x00\x00\x0a\x0b"

    m2 = mc.ArrayMachine(m, 2, add_size_header=False)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, [0x0a, 0x0b])
    assert b.asbytes() == b"\x0a\x0b"

    m2 = mc.ArrayMachine(m, 2, add_size_header=True)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, [0x0a, 0x0b])
    assert b.asbytes() == b"\x02\x00\x00\x00\x0a\x0b"

    m2 = mc.PlainCdrV2SequenceOfPrimitiveMachine(tp.uint8)
    b.seek(0)
    m2.serialize(b, [0x0a, 0x0b])
    assert b.asbytes() == b"\x02\x00\x00\x00\x0a\x0b"

    m2 = mc.PlainCdrV2ArrayOfPrimitiveMachine(tp.uint8, 2)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, [0x0a, 0x0b])
    assert b.asbytes() == b"\x0a\x0b"

    m2 = mc.ByteArrayMachine(3)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, bytearray([1,2,3]))
    m2.serialize(b, b"\x03\x02\x01")
    assert b.asbytes() == b"\x01\x02\x03\x03\x02\x01"

    m2 = mc.EnumMachine(A)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, A.V2)
    m2.serialize(b, A.V1)
    assert b.asbytes() == b"\xef\x00\x00\x00\xca\x00\x00\x00"

    m2 = mc.BitBoundEnumMachine(A, 8)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, A.V2)
    m2.serialize(b, A.V1)
    assert b.asbytes() == b"\xef\xca"

    m2 = mc.BitBoundEnumMachine(A, 16)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, A.V2)
    m2.serialize(b, A.V1)
    assert b.asbytes() == b"\xef\x00\xca\x00"

    m2 = mc.BitMaskMachine(B, 8)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, B(True, False))
    m2.serialize(b, B(False, True))
    assert b.asbytes() == b"\x01\x02"

    m2 = mc.BitMaskMachine(B, 16)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, B(True, False))
    m2.serialize(b, B(False, True))
    assert b.asbytes() == b"\x01\x00\x02\x00"

    m2 = mc.OptionalMachine(m, 0, True)
    b.zero_out()
    b.seek(0)
    m2.serialize(b, None)
    m2.serialize(b, 0x12)
    m2.serialize(b, 0)
    assert b.asbytes() == b"\x00\x01\x12\x01\x00"

    m2 = mc.StringMachine()
    b.zero_out()
    b.seek(0)
    m2.serialize(b, "ab")
    assert b.asbytes() == b"\x03\x00\x00\x00ab\x00"

    m3 = mc.StructMachine(C, {
        'A': mc.PrimitiveMachine(tp.uint8),
        'B': mc.PrimitiveMachine(tp.uint16)
    }, [])
    b.zero_out()
    b.seek(0)
    m3.serialize(b, C(0x0a, 0x0b))
    assert b.asbytes() == b"\x0a\x00\x0b\x00"

    m3 = mc.DelimitedCdrAppendableStructMachine(C, {
        'A': mc.PrimitiveMachine(tp.uint8),
        'B': mc.PrimitiveMachine(tp.uint16)
    }, [])
    b.zero_out()
    b.seek(0)
    m3.serialize(b, C(0x0a, 0x0b))
    assert b.asbytes() == b"\x04\x00\x00\x00\x0a\x00\x0b\x00"

    m3 = mc.PLCdrMutableStructMachine(C, [
        mc.MutableMember(
            'A', key=False, optional=True, lentype=mc.LenType.OneByte,
            must_understand=True, memberid=1, machine=mc.PrimitiveMachine(tp.uint8)
        ),
        mc.MutableMember(
            'B', key=False, optional=True, lentype=mc.LenType.TwoByte,
            must_understand=True, memberid=2, machine=mc.PrimitiveMachine(tp.uint16)
        )
    ], True)

    b.zero_out()
    b.seek(0)
    m3.serialize(b, C(0x0a, 0x0b))
    assert b.asbytes() == (
        b"\x0e\x00\x00\x00"
        b"\x01\x00\x00\x80"
        b"\x0a\x00\x00\x00"
        b"\x02\x00\x00\x90"
        b"\x0b\x00"
    )

    b.zero_out()
    b.seek(0)
    m3.serialize(b, C(None, 0x0b))
    assert b.asbytes() == (
        b"\x06\x00\x00\x00"
        b"\x02\x00\x00\x90"
        b"\x0b\x00"
    )

    b.zero_out()
    b.seek(0)
    m3.serialize(b, C(None, None))
    assert b.asbytes() == b"\x00\x00\x00\x00"

    m3 = mc.UnionMachine(D, mc.PrimitiveMachine(bool), {
        True: mc.PrimitiveMachine(tp.uint16),
    }, default_case=mc.PrimitiveMachine(tp.uint8))

    b.zero_out()
    b.seek(0)
    m3.serialize(b, D(A=0x1234))
    m3.serialize(b, D(B=0x77))
    assert b.asbytes() == b"\x01\x00\x34\x12\x00\x77"

    m3 = mc.DelimitedCdrAppendableUnionMachine(D, mc.PrimitiveMachine(bool), {
        True: mc.PrimitiveMachine(tp.uint16),
    }, default_case=mc.PrimitiveMachine(tp.uint8))

    b.zero_out()
    b.seek(0)
    m3.serialize(b, D(A=0x1234))
    m3.serialize(b, D(B=0x77))
    assert b.asbytes() == b"\x04\x00\x00\x00\x01\x00\x34\x12\x02\x00\x00\x00\x00\x77"







import pytest
import random

from cyclonedds.core import DDSException
from cyclonedds.domain import DomainParticipant
from cyclonedds.topic import Topic
from cyclonedds.pub import Publisher, DataWriter
from cyclonedds.util import duration, isgoodentity
from cyclonedds.sub import DataReader
from cyclonedds.core import Qos, Policy

from support_modules.testtopics import Message, MessageKeyed


def test_initialize_writer():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    pub = Publisher(dp)
    dw = DataWriter(pub, tp)

    assert isgoodentity(dw)


def test_writeto_writer():
    dp = DomainParticipant(0)
    tp = Topic(dp, "Message", Message)
    pub = Publisher(dp)
    dw = DataWriter(pub, tp)

    msg = Message(
        message="TestMessage"
    )

    dw.write(msg)
    assert dw.wait_for_acks(duration(seconds=1))


def test_writer_instance():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageKeyed", MessageKeyed)
    pub = Publisher(dp)
    dw = DataWriter(pub, tp)

    msg = MessageKeyed(user_id=1, message="Hello")

    handle = dw.register_instance(msg)
    assert handle > 0
    dw.write(msg)
    dw.unregister_instance(msg)


def test_writer_instance_handle():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageKeyed", MessageKeyed)
    pub = Publisher(dp)
    dw = DataWriter(pub, tp)

    msg = MessageKeyed(user_id=1, message="Hello")

    handle = dw.register_instance(msg)
    assert handle > 0
    dw.write(msg)
    dw.unregister_instance_handle(handle)


def test_writer_writedispose():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageKeyed", MessageKeyed)
    pub = Publisher(dp)
    dw = DataWriter(pub, tp)

    msg = MessageKeyed(user_id=1, message="Hello")

    dw.write_dispose(msg)


def test_writer_lookup():
    dp = DomainParticipant(0)
    tp = Topic(dp, "MessageKeyed", MessageKeyed)
    pub = Publisher(dp)
    dw = DataWriter(pub, tp)

    keymsg1 = MessageKeyed(user_id=1000, message="Hello!")
    keymsg2 = MessageKeyed(user_id=2000, message="Hello!")
    assert None == dw.lookup_instance(keymsg1)
    assert None == dw.lookup_instance(keymsg2)
    handle1 = dw.register_instance(keymsg1)
    handle2 = dw.register_instance(keymsg2)
    assert handle1 > 0 and handle2 > 0 and handle1 != handle2
    assert handle1 == dw.lookup_instance(keymsg1)
    assert handle2 == dw.lookup_instance(keymsg2)


def test_get_matched_subscriptions():
    dp = DomainParticipant(0)
    tp = Topic(dp, f"Message{random.randint(1000000,9999999)}", Message)
    dw = DataWriter(dp, tp)

    rand_dr = random.randint(0, 20)
    dr = []
    for i in range(rand_dr):
        dr.append(DataReader(dp, tp))

    matched = dw.get_matched_subscriptions()
    assert len(matched) == rand_dr


def test_get_matched_subscription_data():
    dp = DomainParticipant(0)
    tp = Topic(dp, f"Message{random.randint(1000000,9999999)}", Message)
    dr = DataReader(dp, tp)
    dw = DataWriter(dp, tp)

    matched_handles = dw.get_matched_subscriptions()
    for handle in matched_handles:
        matched_data = dw.get_matched_subscription_data(handle)
        print(f"matched data = {matched_data.key}")
        assert matched_data is not None
